import os
import time
import inspect
from osgeo import gdal
from spatialist import bbox, intersect
from spatialist.ancillary import finder
from pyroSAR import identify, identify_many, Archive
from s1ard import etad, ard
from s1ard.config import get_config, gdal_conf
from s1ard.ancillary import set_logging
from s1ard import search
from s1ard import ocn
from cesard import dem
import cesard.tile_extraction as tile_ex
from cesard.search import scene_select
from cesard.ancillary import (buffer_time, check_scene_consistency,
                              check_spacing, get_max_ext, group_by_attr)

from s1ard.processors.registry import load_processor

gdal.UseExceptions()


def main(config_file=None, debug=False, **kwargs):
    """
    Main function that initiates and controls the processing workflow.
    
    Parameters
    ----------
    config_file: str or None
        Full path to a `config.ini` file or `None` to use the package's default file.
    debug: bool
        Set logging level to DEBUG? Default is False.
    **kwargs
        extra arguments to override parameters in the config file. E.g. `acq_mode`.
    """
    update = False  # update existing products? Internal development flag.
    config = get_config(config_file=config_file, **kwargs)
    log = set_logging(config=config, debug=debug)
    config_proc = config['processing']
    processor_name = config_proc['processor']
    processor = load_processor(processor_name)
    config_sar = config[processor_name]
    gdal_prms = gdal_conf(config=config)
    
    spacings = {'IW': 10, 'SM': 10, 'EW': 30}
    config_sar['spacing'] = spacings[config_proc['acq_mode']]
    
    check_spacing(config_sar['spacing'])
    
    sar_flag = 'sar' in config_proc['mode']
    nrb_flag = 'nrb' in config_proc['mode']
    orb_flag = 'orb' in config_proc['mode']
    
    # DEM download authentication
    username, password = dem.authenticate(dem_type=config_proc['dem_type'],
                                          username=None, password=None)
    ####################################################################################################################
    # scene selection
    log.info('collecting scenes')
    
    db_file_set = config_proc['db_file'] is not None
    scene_dir_set = config_proc['scene_dir'] is not None
    stac_catalog_set = config_proc['stac_catalog'] is not None
    stac_collections_set = config_proc['stac_collections'] is not None
    parquet_set = config_proc['parquet'] is not None
    
    if db_file_set:
        archive = Archive(dbfile=config_proc['db_file'])
        if scene_dir_set:
            scenes = finder(target=config_proc['scene_dir'],
                            matchlist=[r'^S1[ABCD].*(SAFE|zip)$'],
                            regex=True, recursive=True, foldermode=1)
            archive.insert(scenes)
    elif stac_catalog_set and stac_collections_set:
        archive = search.STACArchive(url=config_proc['stac_catalog'],
                                     collections=config_proc['stac_collections'])
    elif parquet_set:
        archive = search.STACParquetArchive(files=config_proc['parquet'])
    else:
        raise RuntimeError('could not select a search option. Please check your configuration.')
    
    if config_proc['scene'] is None:
        attr_search = ['sensor', 'product', 'mindate', 'maxdate',
                       'aoi_tiles', 'aoi_geometry', 'date_strict']
        dict_search = {k: config_proc[k] for k in attr_search}
        dict_search['acquisition_mode'] = config_proc['acq_mode']
        
        if config_proc['datatake'] is not None:
            frame_number = [int(x, 16) for x in config_proc['datatake']]
        else:
            frame_number = None
        dict_search['frameNumber'] = frame_number
        
        selection, aoi_tiles = scene_select(archive=archive,
                                            **dict_search)
        
        if len(selection) == 0:
            log.error('could not find any scenes')
            archive.close()
            return
        
        log.info(f'found {len(selection)} scene(s)')
        scenes = identify_many(selection, sortkey='start')
        search.check_acquisition_completeness(scenes=scenes, archive=archive)
    else:
        if config_proc['mode'] != ['sar']:
            raise RuntimeError("if argument 'scene' is set, the processing mode must be 'sar'")
        scenes = [identify(config_proc['scene'])]
        config_proc['acq_mode'] = scenes[0].acquisition_mode
        config_proc['product'] = scenes[0].product
        aoi_tiles = []
    
    # group scenes by datatake
    scenes_grouped = group_by_attr(scenes, lambda x: x.meta['frameNumber'])
    
    for scenes in scenes_grouped:
        # check that the scenes can really be grouped together
        check_scene_consistency(scenes=scenes)
        
        # Remove scenes with an invalid (0) slice number if others have a valid one (>0).
        # This ensures that scenes with a valid slice number are preferred.
        slice_numbers = [x.meta['sliceNumber'] for x in scenes]
        if None not in slice_numbers:
            if min(slice_numbers) == 0 and max(slice_numbers) > 0:
                for i in reversed(range(len(scenes))):
                    if slice_numbers[i] == 0:
                        del scenes[i]
                        del slice_numbers[i]
                for i in range(1, len(scenes)):
                    if slice_numbers[i] != slice_numbers[i - 1] + 1:
                        raise RuntimeError(f"nonconsecutive scene group, "
                                           f"slice numbers: {slice_numbers}")
    ####################################################################################################################
    # Get neighboring GRD scenes to add a buffer to the geocoded scenes.
    # Otherwise, there will be a gap between final geocoded images.
    # Buffering is only possible if the product composition is 'Sliced'
    # (not 'Assembled' or 'Individual') and thus has a sliceNumber attribute.
    if config_proc['product'] == 'GRD' and sar_flag:
        log.info('collecting GRD neighbors')
        neighbors = []
        for scenes in scenes_grouped:
            if scenes[0].meta['sliceNumber'] is not None:
                neighbors_group = []
                for scene in scenes:
                    neighbors_scene = search.collect_neighbors(archive=archive,
                                                               scene=scene)
                    neighbors_group.append(neighbors_scene)
            else:
                neighbors_group = [None] * len(scenes)
            neighbors.append(neighbors_group)
    else:
        neighbors = [[None] * len(x) for x in scenes_grouped]
    ####################################################################################################################
    # OCN scene selection
    if 'wm' in config_proc['annotation']:
        log.info('collecting OCN products')
        scenes_ocn = []
        for scenes in scenes_grouped:
            scenes_ocn_group = []
            for scene in scenes:
                start, stop = buffer_time(scene.start, scene.stop, seconds=2)
                result = archive.select(product='OCN', mindate=start,
                                        maxdate=stop, date_strict=True)
                if len(result) == 1:
                    scenes_ocn_group.append(result[0])
                else:
                    if len(result) == 0:
                        log.error(f'could not find an OCN product for scene {scene.scene}')
                    else:
                        msg = f'found multiple OCN products for scene {scene.scene}:\n'
                        log.error(msg + '\n'.join(result))
                    archive.close()
                    return
            scenes_ocn_group = identify_many(scenes_ocn_group)
            scenes_ocn.append(scenes_ocn_group)
    else:
        scenes_ocn = [[] for scenes in scenes_grouped]
    
    archive.close()
    ####################################################################################################################
    # annotation layer selection
    annotation = config_proc['annotation']
    measurement = config_proc['measurement']
    export_extra = processor.translate_annotation(annotation=annotation,
                                                  measurement=measurement)
    ####################################################################################################################
    # main SAR processing
    if sar_flag:
        for h, scenes in enumerate(scenes_grouped):
            log.info(f'SAR processing of group {h + 1}/{len(scenes_grouped)}')
            for i, scene in enumerate(scenes):
                scene_base = os.path.splitext(os.path.basename(scene.scene))[0]
                out_dir_scene = os.path.join(config_proc['sar_dir'], scene_base)
                tmp_dir_scene = os.path.join(config_proc['tmp_dir'], scene_base)
                
                log.info(f'processing scene {i + 1}/{len(scenes)}: {scene.scene}')
                if os.path.isdir(out_dir_scene) and not update:
                    log.info('Already processed - Skip!')
                    continue
                else:
                    os.makedirs(out_dir_scene, exist_ok=True)
                    os.makedirs(tmp_dir_scene, exist_ok=True)
                ########################################################################################################
                # Preparation of DEM for SAR processing
                dem_prepare_mode = config_sar['dem_prepare_mode']
                if dem_prepare_mode is not None:
                    fname_dem = dem.prepare(scene=scene, dem_type=config_proc['dem_type'],
                                            dir_out=tmp_dir_scene, username=username,
                                            password=password, mode=dem_prepare_mode,
                                            tr=(config_sar['spacing'], config_sar['spacing']))
                else:
                    fname_dem = None
                ########################################################################################################
                # ETAD correction
                if config_proc['etad']:
                    log.info('ETAD correction')
                    scene = etad.process(scene=scene, etad_dir=config_proc['etad_dir'],
                                         out_dir=tmp_dir_scene)
                ########################################################################################################
                # determination of look factors
                if scene.product == 'SLC':
                    rlks = {'IW': 5,
                            'SM': 6,
                            'EW': 3}[config_proc['acq_mode']]
                    rlks *= int(config_sar['spacing'] / 10)
                    azlks = {'IW': 1,
                             'SM': 6,
                             'EW': 1}[config_proc['acq_mode']]
                    azlks *= int(config_sar['spacing'] / 10)
                else:
                    rlks = azlks = None
                ########################################################################################################
                # main processing routine
                start_time = time.time()
                try:
                    log.info('starting SAR processing')
                    proc_args = {'scene': scene.scene,
                                 'outdir': config_proc['sar_dir'],
                                 'measurement': measurement,
                                 'tmpdir': config_proc['tmp_dir'],
                                 'dem': fname_dem,
                                 'neighbors': neighbors[h][i],
                                 'export_extra': export_extra,
                                 'rlks': rlks, 'azlks': azlks}
                    proc_args.update(config_sar)
                    sig = inspect.signature(processor.process)
                    accepted_params = set(sig.parameters.keys())
                    proc_args = {k: v for k, v in proc_args.items() if k in accepted_params}
                    processor.process(**proc_args)
                    t = round((time.time() - start_time), 2)
                    log.info(f'SAR processing finished in {t} seconds')
                except Exception as e:
                    log.error(msg=e)
                    raise
    ####################################################################################################################
    # OCN preparation
    if sum(len(x) for x in scenes_ocn) > 0:
        log.info('extracting OCN products')
        for scenes in scenes_ocn:
            for scene in scenes:
                if scene.compression is not None:
                    scene.unpack(directory=config_proc['tmp_dir'], exist_ok=True)
                basename = os.path.basename(scene.scene).replace('.SAFE', '')
                outdir = os.path.join(config_proc['sar_dir'], basename)
                os.makedirs(outdir, exist_ok=True)
                for v in ['owiNrcsCmod', 'owiEcmwfWindSpeed', 'owiEcmwfWindDirection']:
                    out = os.path.join(outdir, f'{v}.tif')
                    if not os.path.isfile(out):
                        ocn.extract(src=scene.scene, dst=out, variable=v)
    ####################################################################################################################
    # ARD - final product generation
    if nrb_flag or orb_flag:
        product_type = 'NRB' if nrb_flag else 'ORB'
        log.info(f'starting {product_type} production')
        
        for s, scenes in enumerate(scenes_grouped):
            log.info(f'ARD processing of group {s + 1}/{len(scenes_grouped)}')
            log.info('preparing WBM tiles')
            vec = [x.geometry() for x in scenes]
            extent = get_max_ext(geometries=vec)
            with bbox(coordinates=extent, crs=4326) as box:
                dem.retile(vector=box, threads=gdal_prms['threads'],
                           dem_dir=None, wbm_dir=config_proc['wbm_dir'],
                           dem_type=config_proc['dem_type'],
                           tilenames=aoi_tiles, username=username, password=password,
                           dem_strict=True)
            # get the geometries of all tiles that overlap with the current scene group
            tiles = tile_ex.tile_from_aoi(vector=vec,
                                          return_geometries=True,
                                          tilenames=aoi_tiles)
            del vec
            t_total = len(tiles)
            for t, tile in enumerate(tiles):
                # select all scenes from the group whose footprint overlaps with the current tile
                scenes_sub = [x for x in scenes if intersect(tile, x.geometry())]
                scenes_sub_fnames = [x.scene for x in scenes_sub]
                fname_wbm = os.path.join(config_proc['wbm_dir'],
                                         config_proc['dem_type'],
                                         '{}_WBM.tif'.format(tile.mgrs))
                if not os.path.isfile(fname_wbm):
                    fname_wbm = None
                add_dem = True  # add the DEM as output layer?
                dem_type = config_proc['dem_type'] if add_dem else None
                extent = tile.extent
                epsg = tile.getProjection('epsg')
                log.info(f'creating product {t + 1}/{t_total}')
                log.info(f'selected scene(s): {scenes_sub_fnames}')
                try:
                    prod_meta = ard.product_info(product_type=product_type, src_ids=scenes_sub,
                                                 tile_id=tile.mgrs, extent=extent, epsg=epsg,
                                                 dir_ard=config_proc['ard_dir'], update=update)
                except RuntimeError:
                    log.info('Already processed - Skip!')
                    del tiles
                    return
                log.info(f'product name: {prod_meta['dir_ard_product']}')
                try:
                    src_ids, sar_assets = ard.get_datasets(scenes=scenes_sub_fnames,
                                                           sar_dir=config_proc['sar_dir'],
                                                           extent=extent, epsg=epsg,
                                                           processor_name=processor_name)
                    
                    ard_assets = ard.format(config=config, prod_meta=prod_meta,
                                            src_ids=src_ids, sar_assets=sar_assets,
                                            tile=tile.mgrs, extent=extent, epsg=epsg,
                                            wbm=fname_wbm, dem_type=dem_type, compress='LERC_ZSTD',
                                            multithread=gdal_prms['multithread'], annotation=annotation)
                    
                    if ard_assets is not None:
                        ard.append_metadata(config=config, prod_meta=prod_meta,
                                            src_ids=src_ids, assets=ard_assets, compression='LERC_ZSTD')
                
                except Exception as e:
                    log.error(msg=e)
                    raise
            del tiles
        gdal.SetConfigOption('GDAL_NUM_THREADS', gdal_prms['threads_before'])



import os
import re
import time
import shutil
import tarfile as tf
import zipfile as zf
from pyroSAR import identify
from spatialist.ancillary import finder
from s1etad_tools.cli.slc_correct import s1etad_slc_correct_main
import logging

log = logging.getLogger('s1ard')


def process(scene, etad_dir, out_dir):
    """
    Apply ETAD correction to a Sentinel-1 SLC product.
    
    Parameters
    ----------
    scene: pyroSAR.drivers.ID
        The Sentinel-1 SLC scene.
    etad_dir: str
        The directory containing ETAD products. This will be searched for products matching the defined SLC.
    out_dir: str
        The directory to store results. The ETAD product is unpacked to this directory if necessary.
        Two new sub-directories SLC_original SLC_ETAD and are created, which contain the original unpacked
        scene and the corrected one respectively.

    Returns
    -------
    pyroSAR.drivers.ID
        The corrected scene as a pyroSAR ID object.
    """
    slc_corrected_dir = os.path.join(out_dir, 'SLC_etad')
    os.makedirs(slc_corrected_dir, exist_ok=True)
    slc_base = os.path.basename(scene.scene).replace('.zip', '.SAFE')
    slc_corrected = os.path.join(slc_corrected_dir, slc_base)
    if not os.path.isdir(slc_corrected):
        start_time = time.time()
        items = re.match(scene.pattern, os.path.basename(scene.file)).groupdict()
        pattern = '{sensor}_{beam}_ETA__AX{pols}_{start}_{stop}.*(SAFE|zip|tar)$'.format(**items)
        result = finder(etad_dir, [pattern], regex=True, foldermode=1)
        try:
            if len(result) == 0:
                raise RuntimeError('cannot find ETAD product for scene {}'.format(scene.scene))
            match = result[0]
            ext = os.path.splitext(match)[1]
            if ext in ['.tar', '.zip']:
                etad_base = os.path.basename(match).replace(ext, '.SAFE')
                etad = os.path.join(out_dir, etad_base)
                if not os.path.isdir(etad):
                    if ext == '.tar':
                        archive = tf.open(match, 'r')
                    else:
                        archive = zf.ZipFile(match, 'r')
                    archive.extractall(out_dir)
                    archive.close()
            elif ext == '.SAFE':
                etad = match
            else:
                raise RuntimeError('ETAD products are required to be .tar/.zip archives or .SAFE folders')
            scene.unpack(os.path.join(out_dir, 'SLC_original'), exist_ok=True)
            s1etad_slc_correct_main(s1_product=scene.scene,
                                    etad_product=etad,
                                    outdir=slc_corrected_dir,
                                    nthreads=2,
                                    order=0)  # using the default 1 introduces a bias of about -0.5 dB.
            shutil.rmtree(os.path.join(out_dir, 'SLC_original'))
            t = round((time.time() - start_time), 2)
            log.info(f'ETAD correction finished in {t} seconds')
        except Exception as e:
            log.error(e)
            raise
    else:
        msg = 'Already processed - Skip!'
        print('### ' + msg)
    return identify(slc_corrected)



import os
import pytest
import platform


@pytest.fixture(autouse=True)
def tmp_home(monkeypatch, tmp_path):
    home = tmp_path / 'tmp_home'
    home.mkdir()
    var = 'USERPROFILE' if platform.system() == 'Windows' else 'HOME'
    monkeypatch.setenv(var, str(home))
    assert os.path.expanduser('~') == str(home)
    yield home


@pytest.fixture
def stac():
    return {'url': 'https://stac.terrabyte.lrz.de/public/api',
            'collection': 'sentinel-1-grd'}

@pytest.fixture
def stac_parquet():
    return '/dss/dsstbyfs03/pn56su/pn56su-dss-0022/Sentinel-1/GRD/geoparquet/*.parquet'

@pytest.fixture
def testdir():
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')


@pytest.fixture
def testdata(testdir):
    out = {
        's1': os.path.join(testdir, 'S1A_IW_GRDH_1SDV_20200708T182643_20200708T182708_033367_03DDAA_9550.zip'),
    }
    return out



import os
import pytest
from s1ard.config import get_config, init, write
from s1ard.snap import get_config_keys as snap_get_keys


def test_config(tmpdir):
    with pytest.raises(ValueError):
        config = get_config()
    with pytest.raises(ValueError):
        config = get_config(work_dir=str(tmpdir), aoi_tiles='xyz')
    config = get_config(work_dir=str(tmpdir), db_file='scenes.db', aoi_tiles='32TNT')
    assert config['processing']['aoi_tiles'] == ['32TNT']
    assert config['processing']['db_file'] == os.path.join(str(tmpdir), 'scenes.db')


def test_config_snap(tmpdir):
    config = get_config(work_dir=str(tmpdir), db_file='scenes.db', aoi_tiles='32TNT')
    assert 'snap' in config.keys()
    snap_keys = list(config['snap'].keys())
    del snap_keys[snap_keys.index('dem_prepare_mode')]
    assert sorted(snap_keys) == snap_get_keys()
    expected = {
        'allow_res_osv': True,
        'cleanup': True,
        'clean_edges': True,
        'clean_edges_pixels': 4,
        'dem_resampling_method': 'BILINEAR_INTERPOLATION',
        'gpt_args': None,
        'img_resampling_method': 'BILINEAR_INTERPOLATION'
    }
    
    for key, value in expected.items():
        assert config['snap'][key] == value
    
    out = str(tmpdir / 'config.ini')
    write(config, out)
    config = get_config(config_file=out)
    assert 'snap' in config.keys()
    snap_keys = list(config['snap'].keys())
    del snap_keys[snap_keys.index('dem_prepare_mode')]
    assert sorted(snap_keys) == snap_get_keys()
    for key, value in expected.items():
        assert config['snap'][key] == value


def test_init(tmpdir):
    target = str(tmpdir / 'config.ini')
    # work_dir undefined
    with pytest.raises(ValueError):
        init(target=target)
    # no search option defined
    with pytest.raises(RuntimeError):
        init(target=target, work_dir=str(tmpdir))
    gpt_args = ['-J-Xmx100G', '-c', '75G', '-q', '30']
    init(target=target, work_dir=str(tmpdir), db_file='scenes.db',
         gpt_args=gpt_args)
    config = get_config(target)
    assert config['snap']['gpt_args'] == gpt_args
    # file already exists
    with pytest.raises(RuntimeError):
        init(target=target, work_dir=str(tmpdir), db_file='scenes.db')



from __future__ import annotations
from typing import Protocol, Any
from configparser import ConfigParser


class Processor(Protocol):
    """Common interface all processors should follow."""
    name: str
    
    @staticmethod
    def config_to_string(config: dict[str, Any]) -> dict[str, str]: ...
    
    @staticmethod
    def find_datasets(scene: str, outdir: str, epsg: int) -> dict[str, str] | None: ...
    
    @staticmethod
    def get_config_keys() -> list[str]: ...
    
    @staticmethod
    def get_config_section(parser: ConfigParser, **kwargs: Any) -> dict[str, str]: ...
    
    @staticmethod
    def get_metadata(scene: str, outdir: str) -> dict[str, Any]: ...
    
    @staticmethod
    def lsm_encoding() -> dict[str, int]: ...
    
    @staticmethod
    def process(scene: str, outdir: str, **kwargs: Any) -> None: ...
    
    @staticmethod
    def translate_annotation(annotation: list[str] | None, measurement: str) -> list[str]: ...
    
    @staticmethod
    def version_dict() -> dict[str, str]: ...



import os
import re
import shutil
from datetime import datetime, timezone
from copy import deepcopy
from spatialist.vector import Vector, bbox, intersect
from spatialist.raster import Raster
from spatialist.auxil import gdalwarp, gdalbuildvrt
from spatialist.ancillary import finder
from pyroSAR import identify_many
from pyroSAR.ancillary import Lock
from pyroSAR.drivers import ID
import s1ard
from s1ard import ocn
from s1ard.metadata.mapping import ARD_PATTERN
from s1ard.metadata.extract import append_wind_norm, copy_src_meta, meta_dict
from s1ard.processors.registry import load_processor
from cesard.ancillary import (generate_unique_id, vrt_add_overviews,
                              datamask, get_tmp_name)
from cesard.ard import (calc_product_start_stop, create_acq_id_image,
                        create_data_mask, create_rgb_vrt, create_vrt)
from cesard import dem
from cesard.metadata import xml, stac
from cesard.metadata.mapping import LERC_ERR_THRES
import logging

log = logging.getLogger('s1ard')


def product_info(
        product_type: str,
        src_ids: list[ID],
        tile_id: str,
        extent: dict[str, int | float],
        epsg: int,
        dir_ard: str,
        update: bool = False,
        product_id: str | None = None
) -> dict[str, str | int | datetime]:
    """
    Create ARD product metadata.
    
    Parameters
    ----------
    product_type:
        the ARD product type; options: NRB, ORB
    src_ids:
        the source product objects
    tile_id:
        the MGRS tile ID
    extent:
        the extent of the MGRS tile
    epsg:
        the EPSG code of the MGRS tile
    dir_ard:
        the ARD output directory
    update:
        update an existing product (or create a new one)
    product_id:
        an existing product ID. Default None: create a new one using
        function :func:`generate_unique_id`.

    Returns
    -------
        ARD product metadata
    """
    # determine processing timestamp and generate unique ID
    proc_time = datetime.now(timezone.utc)
    if product_id is None:
        t = proc_time.isoformat().encode()
        product_id = generate_unique_id(encoded_str=t)
    
    ard_start, ard_stop = calc_product_start_stop(src_ids=src_ids, extent=extent, epsg=epsg)
    pol_str = '_'.join(sorted(src_ids[0].polarizations))
    meta = {'mission': src_ids[0].sensor,
            'mode': src_ids[0].meta['acquisition_mode'],
            'product_type': product_type,
            'polarization': {'HH': 'SH',
                             'VV': 'SV',
                             'HH_HV': 'DH',
                             'VH_VV': 'DV'}[pol_str],
            'start': ard_start,
            'stop': ard_stop,
            'proc_time': proc_time,
            'orbitnumber': src_ids[0].meta['orbitNumbers_abs']['start'],
            'datatake': hex(src_ids[0].meta['frameNumber']).replace('x', '').upper(),
            'tile': tile_id,
            'id': product_id}
    meta_name = deepcopy(meta)
    meta_name['start'] = datetime.strftime(meta_name['start'], '%Y%m%dT%H%M%S')
    del meta_name['stop']
    del meta_name['proc_time']
    meta_name_lower = dict((k, v.lower() if isinstance(v, str) else v)
                           for k, v in meta_name.items())
    skeleton_dir = ('{mission}_{mode}_{product_type}__1S{polarization}_{start}_'
                    '{orbitnumber:06}_{datatake:0>6}_{tile}_{id}')
    skeleton_files = '{mission}-{mode}-{product_type}-{start}-{orbitnumber:06}-{datatake:0>6}-{tile}'
    
    meta['product_base'] = skeleton_dir.format(**meta_name)
    dir_ard_tile = os.path.join(dir_ard, tile_id)
    os.makedirs(dir_ard_tile, exist_ok=True)
    meta['dir_ard_product'] = os.path.join(dir_ard_tile, meta['product_base'])
    meta['file_base'] = skeleton_files.format(**meta_name_lower) + '-{suffix}.tif'
    
    # check existence of products
    msg = 'Already processed - Skip!'
    pattern = meta['product_base'].replace(product_id, '*')
    existing = finder(dir_ard_tile, [pattern], foldermode=2)
    if len(existing) > 0:
        if not update:
            raise RuntimeError(msg)
        else:
            if existing[0] != meta['dir_ard_product']:
                existing_meta = re.search(ARD_PATTERN, os.path.basename(existing[0])).groupdict()
                return product_info(product_type=product_type, src_ids=src_ids,
                                    tile_id=tile_id, extent=extent, epsg=epsg,
                                    dir_ard=dir_ard, update=update,
                                    product_id=existing_meta['ID'])
            else:
                return meta
    else:
        try:
            os.makedirs(meta['dir_ard_product'], exist_ok=False)
        except OSError:
            raise RuntimeError(msg)
    return meta


def format(
        config: dict[str, dict[str, int | float | str | list[str]]],
        prod_meta: dict[str, str | int | datetime],
        src_ids: list[ID],
        sar_assets: list[dict[str, str]],
        tile: str,
        extent: dict[str, int | float],
        epsg: int,
        wbm: str | None = None,
        dem_type: str | None = None,
        multithread: bool = True,
        compress: str | None = None,
        overviews: list[int] | None = None,
        annotation: list[str] | None = None
) -> list[str] | None:
    """
    Create ARD products from the SAR processor output.
    This includes the following:
    
    - Creating all measurement and annotation datasets in Cloud Optimized GeoTIFF (COG) format
    - Creating additional annotation datasets in Virtual Raster Tile (VRT) format
    - Applying the ARD product directory structure & naming convention
    - Generating metadata in XML and JSON formats for the ARD product as well as source SLC datasets
    
    Parameters
    ----------
    config:
        Dictionary of the parsed config parameters for the current process.
    prod_meta:
        Product metadata as returned by :func:`~s1ard.ard.product_info`.
    src_ids:
        List of scenes to process. Either a single scene or multiple, matching scenes (consecutive acquisitions).
        All scenes are expected to overlap with `extent` and an error will be thrown if the processing output
        cannot be found for any of the scenes.
    sar_assets:
        The SAR processing assets as returned by :func:`get_datasets`.
    tile:
        ID of an MGRS tile.
    extent:
        Spatial extent of the MGRS tile, derived from a :class:`~spatialist.vector.Vector` object.
    epsg:
        The CRS used for the ARD product; provided as an EPSG code.
    wbm:
        Path to a water body mask file with the dimensions of an MGRS tile.
    dem_type:
        if defined, a DEM layer will be added to the product. The suffix `em` (elevation model) is used.
        Default `None`: do not add a DEM layer.
    multithread:
        Should `gdalwarp` use multithreading? Default is True. The number of threads used, can be adjusted in the
        `config.ini` file with the parameter `gdal_threads`.
    compress:
        Compression algorithm to use. See https://gdal.org/drivers/raster/gtiff.html#creation-options for options.
        Defaults to 'LERC_DEFLATE'.
    overviews:
        Internal overview levels to be created for each GeoTIFF file. Defaults to [2, 4, 9, 18, 36]
    annotation:
        an optional list to select the annotation layers. Default `None`: create all layers if the
        source products contain the required input layers. Options:
        
        - dm: data mask (four masks: not layover not shadow, layover, shadow, water)
        - ei: ellipsoidal incident angle
        - em: digital elevation model
        - id: acquisition ID image (source scene ID per pixel)
        - lc: RTC local contributing area
        - ld: range look direction angle
        - li: local incident angle
        - np: noise power (NESZ, per polarization)
        - gs: gamma-sigma ratio: sigma0 RTC / gamma0 RTC
        - sg: sigma-gamma ratio: gamma0 RTC / sigma0 ellipsoidal
        - wm: OCN product wind model; requires OCN scenes via argument `scenes_ocn`
    
    Returns
    -------
        the ARD product assets
    """
    if compress is None:
        compress = 'LERC_ZSTD'
    if overviews is None:
        overviews = [2, 4, 9, 18, 36]
    ovr_resampling = 'AVERAGE'
    driver = 'COG'
    blocksize = 512
    dst_nodata_float = -9999.0
    dst_nodata_byte = 255
    vrt_nodata = 'nan'  # was found necessary for proper calculation of statistics in QGIS
    vrt_options = {'VRTNodata': vrt_nodata}
    
    processor_name = config['processing']['processor']
    
    if len(src_ids) == 0:
        log.error(f'None of the processed scenes overlap with the current tile {tile}')
        shutil.rmtree(prod_meta['dir_ard_product'])
        return
    
    if annotation is not None:
        allowed = []
        for key in sar_assets[0]:
            c1 = re.search('[gs]-lin', key)
            c2 = key in annotation
            c3 = key in ['gs', 'sg'] and 'ratio' in annotation
            c4 = key.startswith('np') and 'np' in annotation
            if c1 or c2 or c3 or c4:
                allowed.append(key)
    else:
        allowed = [key for key in sar_assets[0].keys() if re.search('[gs]-lin', key)]
        annotation = []
    for item in ['em', 'id']:
        if item in annotation:
            allowed.append(item)
    
    # GDAL output bounds
    bounds = [extent['xmin'], extent['ymin'], extent['xmax'], extent['ymax']]
    
    subdirectories = ['measurement', 'annotation', 'source', 'support']
    for subdirectory in subdirectories:
        os.makedirs(os.path.join(prod_meta['dir_ard_product'], subdirectory), exist_ok=True)
    
    # prepare raster write options; https://gdal.org/drivers/raster/cog.html
    write_options_base = ['BLOCKSIZE={}'.format(blocksize),
                          'OVERVIEW_RESAMPLING={}'.format(ovr_resampling)]
    write_options = dict()
    for key in LERC_ERR_THRES:
        write_options[key] = write_options_base.copy()
        if compress is not None:
            entry = 'COMPRESS={}'.format(compress)
            write_options[key].append(entry)
            if compress.startswith('LERC'):
                entry = 'MAX_Z_ERROR={:f}'.format(LERC_ERR_THRES[key])
                write_options[key].append(entry)
    
    # create raster files: linear gamma0/sigma0 backscatter (-[vh|vv|hh|hv]-[gs]-lin.tif),
    # ellipsoidal incident angle (-ei.tif), gamma-to-sigma ratio (-gs.tif),
    # local contributing area (-lc.tif), local incident angle (-li.tif),
    # noise power images (-np-[vh|vv|hh|hv].tif)
    ard_assets = dict()
    for key in list(sar_assets[0].keys()):
        if key in ['dm', 'wm'] or key not in LERC_ERR_THRES.keys() or key not in allowed:
            # raster files for keys 'dm' and 'wm' are created later
            continue
        
        outname_base = prod_meta["file_base"].format(suffix=key)
        if re.search('[gs]-lin', key):
            subdir = 'measurement'
        else:
            subdir = 'annotation'
        outname = os.path.join(prod_meta['dir_ard_product'], subdir, outname_base)
        
        if not os.path.isfile(outname):
            log.info(f"creating {os.path.relpath(outname, prod_meta['dir_ard_product'])}")
            images = [ds[key] for ds in sar_assets]
            ras = None
            if len(images) > 1:
                ras = Raster(images, list_separate=False)
                source = ras.filename
            else:
                source = get_tmp_name(suffix='.vrt')
                gdalbuildvrt(src=images[0], dst=source)
            
            # modify temporary VRT to make sure overview levels and resampling are properly applied
            vrt_add_overviews(vrt=source, overviews=overviews, resampling=ovr_resampling)
            
            options = {'format': driver, 'outputBounds': bounds,
                       'dstNodata': dst_nodata_float, 'multithread': multithread,
                       'creationOptions': write_options[key]}
            
            gdalwarp(src=source, dst=outname, **options)
            if ras is not None:
                ras.close()
        ard_assets[key] = outname
    
    # define a reference raster and list all gamma0/sigma0 backscatter measurement rasters
    measure_tifs = [v for k, v in ard_assets.items() if re.search('[gs]-lin', k)]
    ref_key = list(ard_assets.keys())[0]
    ref_tif = ard_assets[ref_key]
    
    # create data mask raster (-dm.tif)
    if 'dm' in allowed:
        if wbm is not None:
            if not config['processing']['dem_type'] == 'GETASSE30' and not os.path.isfile(wbm):
                raise FileNotFoundError('External water body mask could not be found: {}'.format(wbm))
        
        dm_path_base = prod_meta["file_base"].format(suffix='dm')
        dm_path = os.path.join(prod_meta['dir_ard_product'], 'annotation', dm_path_base)
        if not os.path.isfile(dm_path):
            log.info(f"creating {os.path.relpath(dm_path, prod_meta['dir_ard_product'])}")
            processor = load_processor(processor_name)
            lsm_encoding = processor.lsm_encoding()
            create_data_mask(outname=dm_path, datasets=sar_assets, extent=extent, epsg=epsg,
                             driver=driver, creation_opt=write_options['dm'],
                             overviews=overviews, overview_resampling=ovr_resampling,
                             dst_nodata=dst_nodata_byte, wbm=wbm,
                             product_type=prod_meta['product_type'],
                             lsm_encoding=lsm_encoding)
        ard_assets['dm'] = dm_path
    
    # create acquisition ID image raster (-id.tif)
    if 'id' in allowed:
        id_path_base = prod_meta["file_base"].format(suffix='id')
        id_path = os.path.join(prod_meta['dir_ard_product'], 'annotation', id_path_base)
        if not os.path.isfile(id_path):
            log.info(f"creating {os.path.relpath(id_path, prod_meta['dir_ard_product'])}")
            create_acq_id_image(outname=id_path, ref_tif=ref_tif,
                                datasets=sar_assets, src_ids=src_ids,
                                extent=extent, epsg=epsg, driver=driver,
                                creation_opt=write_options['id'],
                                overviews=overviews, dst_nodata=dst_nodata_byte)
        ard_assets['id'] = id_path
    
    # create DEM (-em.tif)
    # (if not already converted from processor output)
    if dem_type is not None and 'em' in allowed:
        em_path_base = prod_meta["file_base"].format(suffix='em')
        em_path = os.path.join(prod_meta['dir_ard_product'], 'annotation', em_path_base)
        if not os.path.isfile(em_path):
            log.info(f"creating {os.path.relpath(em_path, prod_meta['dir_ard_product'])}")
            with Raster(ref_tif) as ras:
                tr = ras.res
            log_pyro = logging.getLogger('pyroSAR')
            level = log_pyro.level
            log_pyro.setLevel('NOTSET')
            dem.to_mgrs(dem_type=dem_type, dst=em_path,
                        overviews=overviews, tile=tile, tr=tr,
                        create_options=write_options['em'],
                        pbar=False)
            log_pyro.setLevel(level)
        ard_assets['em'] = em_path
    
    # create color composite VRT (-cc-[gs]-lin.vrt)
    if prod_meta['polarization'] in ['DH', 'DV'] and len(measure_tifs) == 2:
        cc_path = re.sub('[hv]{2}', 'cc', measure_tifs[0]).replace('.tif', '.vrt')
        if not os.path.isfile(cc_path):
            log.info(f"creating {os.path.relpath(cc_path, prod_meta['dir_ard_product'])}")
            create_rgb_vrt(outname=cc_path, infiles=measure_tifs,
                           overviews=overviews, overview_resampling=ovr_resampling)
        key = re.search('cc-[gs]-lin', cc_path).group()
        ard_assets[key] = cc_path
    
    # create log-scaled gamma0|sigma0 nought VRTs (-[vh|vv|hh|hv]-[gs]-log.vrt)
    fun = 'dB'
    args = {'fact': 10}
    scale = None
    for item in measure_tifs:
        target = item.replace('lin.tif', 'log.vrt')
        if not os.path.isfile(target):
            log.info(f"creating {os.path.relpath(target, prod_meta['dir_ard_product'])}")
            create_vrt(src=item, dst=target, fun=fun, scale=scale,
                       args=args, options=vrt_options, overviews=overviews,
                       overview_resampling=ovr_resampling)
        key = re.search('[hv]{2}-[gs]-log', target).group()
        ard_assets[key] = target
    
    # create sigma nought RTC VRTs (-[vh|vv|hh|hv]-s-[lin|log].vrt)
    if 'gs' in allowed:
        gs_path = ard_assets['gs']
        for item in measure_tifs:
            sigma0_rtc_lin = item.replace('g-lin.tif', 's-lin.vrt')
            sigma0_rtc_log = item.replace('g-lin.tif', 's-log.vrt')
            
            if not os.path.isfile(sigma0_rtc_lin):
                log.info(f"creating {os.path.relpath(sigma0_rtc_lin, prod_meta['dir_ard_product'])}")
                create_vrt(src=[item, gs_path], dst=sigma0_rtc_lin, fun='mul',
                           relpaths=True, options=vrt_options, overviews=overviews,
                           overview_resampling=ovr_resampling)
            key = re.search('[hv]{2}-s-lin', sigma0_rtc_lin).group()
            ard_assets[key] = sigma0_rtc_lin
            
            if not os.path.isfile(sigma0_rtc_log):
                log.info(f"creating {os.path.relpath(sigma0_rtc_log, prod_meta['dir_ard_product'])}")
                create_vrt(src=sigma0_rtc_lin, dst=sigma0_rtc_log, fun=fun,
                           scale=scale, options=vrt_options, overviews=overviews,
                           overview_resampling=ovr_resampling, args=args)
            key = key.replace('lin', 'log')
            ard_assets[key] = sigma0_rtc_log
    
    # create gamma nought RTC VRTs (-[vh|vv|hh|hv]-g-[lin|log].vrt)
    if 'sg' in allowed:
        sg_path = ard_assets['sg']
        for item in measure_tifs:
            if not item.endswith('s-lin.tif'):
                continue
            gamma0_rtc_lin = item.replace('s-lin.tif', 'g-lin.vrt')
            gamma0_rtc_log = item.replace('s-lin.tif', 'g-log.vrt')
            
            if not os.path.isfile(gamma0_rtc_lin):
                log.info(f"creating {os.path.relpath(gamma0_rtc_lin, prod_meta['dir_ard_product'])}")
                create_vrt(src=[item, sg_path], dst=gamma0_rtc_lin, fun='mul',
                           relpaths=True, options=vrt_options, overviews=overviews,
                           overview_resampling=ovr_resampling)
            key = re.search('[hv]{2}-g-lin', gamma0_rtc_lin).group()
            ard_assets[key] = gamma0_rtc_lin
            
            if not os.path.isfile(gamma0_rtc_log):
                log.info(f"creating {os.path.relpath(gamma0_rtc_log, prod_meta['dir_ard_product'])}")
                create_vrt(src=gamma0_rtc_lin, dst=gamma0_rtc_log, fun=fun,
                           scale=scale, options=vrt_options, overviews=overviews,
                           overview_resampling=ovr_resampling, args=args)
            key = key.replace('lin', 'log')
            ard_assets[key] = gamma0_rtc_log
    
    # create backscatter wind model (-wm.tif)
    # and wind normalization VRT (-[vv|hh]-s-lin-wn.vrt)
    wm_ref_speed = wm_ref_direction = None
    if 'wm' in annotation:
        wm = []
        wm_ref_speed = []
        wm_ref_direction = []
        for i, ds in enumerate(sar_assets):
            if 'wm' in ds.keys():
                wm.append(ds['wm'])
                # for key in ['wm_ref_speed', 'wm_ref_direction']:
                if 'wm_ref_speed' in ds.keys():
                    wm_ref_speed.append(ds['wm_ref_speed'])
                if 'wm_ref_direction' in ds.keys():
                    wm_ref_direction.append(ds['wm_ref_direction'])
            else:
                scene_base = os.path.basename(src_ids[i].scene)
                raise RuntimeError(f'could not find wind model product for scene {scene_base}')
        
        copol = 'VV' if 'VV' in src_ids[0].polarizations else 'HH'
        copol_sigma0_key = f'{copol.lower()}-s-lin'
        if copol_sigma0_key in ard_assets.keys():
            copol_sigma0 = ard_assets[copol_sigma0_key]
            wn_ard = re.sub(r's-lin\.(?:tif|vrt)', 's-lin-wn.vrt', copol_sigma0)
        else:
            copol_sigma0 = None
            wn_ard = None
        
        outname_base = prod_meta["file_base"].format(suffix='wm')
        wm_ard = os.path.join(prod_meta['dir_ard_product'], 'annotation', outname_base)
        
        gapfill = True if src_ids[0].product == 'GRD' else False
        
        log.info(f"creating {os.path.relpath(wm_ard, prod_meta['dir_ard_product'])}")
        if wn_ard is not None:
            log.info(f"creating {os.path.relpath(wn_ard, prod_meta['dir_ard_product'])}")
        
        wind_normalization(src=wm, dst_wm=wm_ard, dst_wn=wn_ard, measurement=copol_sigma0,
                           gapfill=gapfill, bounds=bounds, epsg=epsg, driver=driver,
                           creation_opt=write_options['wm'],
                           dst_nodata=dst_nodata_float, multithread=multithread)
        ard_assets['wm'] = wm_ard
        ard_assets[f'{copol_sigma0_key}-wn'] = wn_ard
    
    ard_assets = sorted(sorted(list(ard_assets.values()), key=lambda x: os.path.splitext(x)[1]),
                        key=lambda x: os.path.basename(os.path.dirname(x)), reverse=True)
    
    return ard_assets


def append_metadata(
        config: dict[str, dict[str, int | float | str | list[str]]],
        prod_meta: dict[str, str | int | datetime],
        src_ids: list[ID],
        assets: list[str],
        compression: str,
        wm_ref_speed: list[str] | None = None,
        wm_ref_direction: list[str] | None = None
) -> None:
    """
    Append metadata files to an ARD product.
    
    Parameters
    ----------
    config:
        the configuration dictionary
    prod_meta:
        the product metadata as returned by :func:`product_info`
    src_ids:
        the source product objects
    assets:
        a list of assets in the ARD product as returned by :func:`format`.
    compression:
        the used compression algorithm
    wm_ref_speed:
        the wind model reference wind speed files
    wm_ref_direction:
        the wind model reference wind direction files

    Returns
    -------

    """
    meta = meta_dict(config=config, prod_meta=prod_meta,
                     src_ids=src_ids, compression=compression)
    
    append_wind_norm(meta=meta, wm_ref_speed=wm_ref_speed,
                     wm_ref_direction=wm_ref_direction)
    
    # copy support files
    schema_dir = os.path.join(s1ard.__path__[0], 'validation', 'schemas')
    schemas = os.listdir(schema_dir)
    for schema in schemas:
        schema_in = os.path.join(schema_dir, schema)
        schema_out = os.path.join(prod_meta['dir_ard_product'], 'support', schema)
        if not os.path.isfile(schema_out):
            log.info(f"creating {os.path.relpath(schema_out, prod_meta['dir_ard_product'])}")
            shutil.copy(schema_in, schema_out)
    
    if config['metadata']['copy_original']:
        copy_src_meta(ard_dir=prod_meta['dir_ard_product'], src_ids=src_ids)
    if 'OGC' in config['metadata']['format']:
        xml.parse(meta=meta, target=prod_meta['dir_ard_product'],
                  assets=assets, exist_ok=True)
    if 'STAC' in config['metadata']['format']:
        stac.parse(meta=meta, target=prod_meta['dir_ard_product'],
                   assets=assets, exist_ok=True)


def get_datasets(
        scenes: list[str],
        sar_dir: str,
        extent: dict[str, int | float],
        epsg: int,
        processor_name: str
) -> tuple[list[ID], list[dict[str, str]]]:
    """
    Collect processing output for a list of scenes. Reads metadata from all
    source SLC/GRD scenes, finds matching output files in `sar_dir` and
    filters both lists depending on the actual overlap of each SLC/GRD valid
    data coverage with the current MGRS tile geometry. If no output is found
    for any scene the function will raise an error.
    To obtain the extent of valid data coverage, first a binary mask raster
    file is created with the name `datamask.tif`, which is stored in the same
    folder as the processing output as found by :func:`~s1ard.snap.find_datasets`.
    Then, the boundary of this binary mask is computed and stored as `datamask.gpkg`
    (see function :func:`spatialist.vector.boundary`).
    If the provided `extent` does not overlap with this boundary, the output is
    discarded. This scenario might occur when the scene's geometry read from its
    metadata overlaps with the tile but the actual extent of data does not.

    Parameters
    ----------
    scenes:
        List of scenes to process. Either an individual scene or multiple,
        matching scenes (consecutive acquisitions).
    sar_dir:
        The directory containing the SAR datasets processed from the source
        scenes using pyroSAR. The function will raise an error if the processing
        output cannot be found for all scenes in `sar_dir`.
    extent:
        Spatial extent of the MGRS tile, derived from a
        :class:`~spatialist.vector.Vector` object.
    epsg:
        The coordinate reference system as an EPSG code.
    processor_name:
        The name of the used SAR processor. The function `find_datasets` of the
        respective processor module is used.

    Returns
    -------
        List of :class:`~pyroSAR.drivers.ID` objects of all source SLC/GRD scenes
        that overlap with the current MGRS tile and a list of SAR processing output
        files that match each :class:`~pyroSAR.drivers.ID` object of `ids`.
        The format of the latter is a list of dictionaries per scene with keys as
        described by e.g. :func:`s1ard.snap.find_datasets`.
    
    See Also
    --------
    :func:`s1ard.snap.find_datasets`
    """
    processor = load_processor(processor_name)
    ids = identify_many(scenes, sortkey='start')
    datasets = []
    for i, _id in enumerate(ids):
        log.debug(f'collecting processing output for scene {os.path.basename(_id.scene)}')
        files = processor.find_datasets(scene=_id.scene, outdir=sar_dir, epsg=epsg)
        if files is not None:
            base = os.path.splitext(os.path.basename(_id.scene))[0]
            ocn = re.sub('(?:SLC_|GRD[FHM])_1', 'OCN__2', base)[:-5]
            # allow 1 second tolerance
            s_start = int(ocn[31])
            s_stop = int(ocn[47])
            ocn_list = list(ocn)
            s = 1
            ocn_list[31] = f'[{s_start - s}{s_start}{s_start + s}]'
            ocn_list[47] = f'[{s_stop - s}{s_stop}{s_stop + s}]'
            ocn = ''.join(ocn_list)
            log.debug(f'searching for OCN products with pattern {ocn}')
            ocn_match = finder(target=sar_dir, matchlist=[ocn], regex=True,
                               foldermode=2, recursive=False)
            if len(ocn_match) > 0:
                for v in ['owiNrcsCmod', 'owiEcmwfWindSpeed', 'owiEcmwfWindDirection']:
                    ocn_tif = os.path.join(ocn_match[0], f'{v}.tif')
                    if os.path.isfile(ocn_tif):
                        if v.endswith('Speed'):
                            files['wm_ref_speed'] = ocn_tif
                        elif v.endswith('Direction'):
                            files['wm_ref_direction'] = ocn_tif
                        else:
                            files['wm'] = ocn_tif
            datasets.append(files)
        else:
            base = os.path.basename(_id.scene)
            msg = f'cannot find processing output for scene {base} and CRS EPSG:{epsg}'
            raise RuntimeError(msg)
    
    i = 0
    while i < len(datasets):
        log.debug(f'checking tile overlap for scene {os.path.basename(ids[i].scene)}')
        measurements = [datasets[i][x] for x in datasets[i].keys()
                        if re.search('[gs]-lin', x)]
        dm_ras = os.path.join(os.path.dirname(measurements[0]), 'datamask.tif')
        dm_vec = dm_ras.replace('.tif', '.gpkg')
        dm_vec = datamask(measurement=measurements[0], dm_ras=dm_ras, dm_vec=dm_vec)
        if dm_vec is None:
            del ids[i], datasets[i]
            continue
        with Lock(dm_vec, soft=True):
            with Vector(dm_vec) as bounds:
                with bbox(extent, epsg) as tile_geom:
                    inter = intersect(bounds, tile_geom)
                    if inter is not None:
                        with Raster(dm_ras) as ras:
                            inter_min = ras.res[0] * ras.res[1]
                        if inter.getArea() < inter_min:
                            inter.close()
                            inter = None
                    if inter is None:
                        log.debug('no overlap, removing scene')
                        del ids[i]
                        del datasets[i]
                    else:
                        log.debug('overlap detected')
                        # Add dm_ras to the datasets if it overlaps with the current tile
                        datasets[i]['datamask'] = dm_ras
                        i += 1
                        inter.close()
    return ids, datasets


def wind_normalization(
        src: list[str],
        dst_wm: str,
        dst_wn: str | None,
        measurement: str | None,
        gapfill: bool,
        bounds: list[float],
        epsg: int,
        driver: str,
        creation_opt: list[str],
        dst_nodata: float,
        multithread: bool,
        resolution: int = 915
) -> None:
    """
    Create ORB wind normalization layers. A wind model annotation layer is
    created and optionally a wind normalization VRT.
    
    Parameters
    ----------
    src:
        A list of OCN products as prepared by :func:`s1ard.ocn.extract`
    dst_wm:
        The name of the wind model layer in the ARD product
    dst_wn:
        The name of the wind normalization VRT. If None, no VRT will be created.
        Requires `measurement` to point to a file.
    measurement:
        The name of the measurement file used for wind normalization in `dst_wn`.
        If None, no wind normalization VRT will be created.
    gapfill:
        Perform additional gap filling (:func:`s1ard.ocn.gapfill`)?
        This is recommended if the Level-1 source product of `measurement` is GRD
        in which case gaps are introduced between subsequently acquired scenes.
    bounds:
        the bounds of the MGRS tile
    epsg:
        The EPSG code of the MGRS tile
    driver:
        GDAL driver to use for raster file creation.
    creation_opt:
        GDAL creation options to use for raster file creation. Should match
        specified GDAL driver.
    dst_nodata:
        Nodata value to write to the output raster.
    multithread:
        Should `gdalwarp` use multithreading?
    resolution:
        The target pixel resolution in meters. 915 is chosen as default because
        it is closest to the OCN product resolution (1000) and still fits into
        the MGRS bounds (``109800 % 915 == 0``).
    
    Returns
    -------
    
    """
    if len(src) > 1:
        cmod_mosaic = get_tmp_name(suffix='.tif')
        gdalwarp(src=src, dst=cmod_mosaic)
        if gapfill:
            cmod_geo = get_tmp_name(suffix='.tif')
            ocn.gapfill(src=cmod_mosaic, dst=cmod_geo, md=2, si=1)
            os.remove(cmod_mosaic)
        else:
            cmod_geo = cmod_mosaic
        cmod_geo_tmp = True
    else:
        cmod_geo = src[0]
        cmod_geo_tmp = False
    
    if not os.path.isfile(dst_wm):
        log.info(f'creating {dst_wm}')
        gdalwarp(src=cmod_geo,
                 dst=dst_wm,
                 outputBounds=bounds,
                 dstSRS=f'EPSG:{epsg}',
                 xRes=resolution, yRes=resolution,
                 resampleAlg='bilinear',
                 format=driver,
                 dstNodata=dst_nodata,
                 multithread=multithread,
                 creationOptions=creation_opt)
    
    if cmod_geo_tmp:
        os.remove(cmod_geo)
    
    if dst_wn is not None and measurement is not None:
        if not os.path.isfile(dst_wn):
            log.info(f'creating {dst_wn}')
            with Raster(measurement) as ras:
                xres, yres = ras.res
            create_vrt(src=[measurement, dst_wm], dst=dst_wn, fun='div',
                       options={'xRes': xres, 'yRes': yres}, relpaths=True)



from __future__ import annotations
from typing import Any
from .base import Processor
from configparser import ConfigParser

from s1ard import snap


class SnapProcessor(Processor):
    """Adapter that exposes s1ard.snap functions as instance methods."""
    name = "snap"
    
    @staticmethod
    def config_to_string(config: dict[str, Any]) -> dict[str, str]:
        return snap.config_to_string(config)
    
    @staticmethod
    def find_datasets(scene: str, outdir: str, epsg: int) -> dict[str, str] | None:
        return snap.find_datasets(scene, outdir, epsg)
    
    @staticmethod
    def get_config_keys() -> list[str]:
        return snap.get_config_keys()
    
    @staticmethod
    def get_config_section(parser: ConfigParser, **kwargs: dict[str, str]) -> dict[str, str]:
        return snap.get_config_section(parser, **kwargs)
    
    @staticmethod
    def get_metadata(scene: str, outdir: str) -> dict[str, Any]:
        return snap.get_metadata(scene, outdir)
    
    @staticmethod
    def lsm_encoding() -> dict[str, int]:
        return snap.lsm_encoding()
    
    @staticmethod
    def process(scene: str, outdir: str, tmpdir: str, dem: str, measurement: str,
                spacing: int | float, dem_resampling_method: str,
                img_resampling_method: str, rlks: int | None, azlks: int | None,
                export_extra: list[str] | None, allow_res_osv: bool, clean_edges: bool,
                clean_edges_pixels: int, neighbors: list[str] | None,
                gpt_args: list[str] | None, cleanup: bool) -> None:
        return snap.process(scene=scene, outdir=outdir, tmpdir=tmpdir, dem=dem,
                            measurement=measurement, spacing=spacing,
                            dem_resampling_method=dem_resampling_method,
                            img_resampling_method=img_resampling_method,
                            rlks=rlks, azlks=azlks, export_extra=export_extra,
                            allow_res_osv=allow_res_osv, clean_edges=clean_edges,
                            clean_edges_pixels=clean_edges_pixels,
                            neighbors=neighbors, gpt_args=gpt_args, cleanup=cleanup)
    
    @staticmethod
    def translate_annotation(annotation: list[str], measurement: str) -> list[str]:
        return snap.translate_annotation(annotation, measurement)
    
    @staticmethod
    def version_dict() -> dict[str, str]:
        return snap.version_dict()



import click


@click.group(name='s1rb',
             no_args_is_help=True,
             invoke_without_command=True)
@click.option('--version', is_flag=True,
              help='Print s1ard version information and exit. Overrides all other arguments.')
def cli(version):
    if version:
        import s1ard
        print(s1ard.__version__)


@cli.command(name='init',
             no_args_is_help=True,
             context_settings=dict(
                 ignore_unknown_options=True,
                 allow_extra_args=True, )
             )
@click.option('--config-file', '-c', required=True, type=click.Path(),
              help="Full path to an INI-style target configuration text file.")
@click.option('--overwrite', '-o', is_flag=True, default=False,
              help='Overwrite an existing file?')
@click.option('--config-source', '-s', required=False, type=click.Path(),
              help="Full path to an INI-style source configuration text file. "
                   "If not defined, configuration will be read from the package's default file.")
@click.pass_context
def init(ctx, config_file, overwrite=False, config_source=None):
    """
    Initialize configuration for s1ard radar backscatter processing.

    This creates a processor configuration file in a user-defined location.
    The package's default file or a user-defined source file may serve as base.
    Additional options can be passed to override individual processing parameters
    in the config file. For example, to read all values from the default
    file except the acquisition mode and the annotation layers:

    s1rb init -c config.ini --acq_mode IW --annotation dm,id
    
    A previously created config file can be used instead of the default file:
    
    s1rb init -c config.ini -s config_source.ini --acq_mode IW --annotation dm,id

    SAR processor command line arguments (like SNAP gpt_args) can be provided
    by using a dedicated argument separator and quotes:

    s1rb init -c config.ini -- --gpt_args "-J-Xmx100G -c 75G -q 30"

    \b
    The following defaults are set:
    PROCESSING
    - annotation:        dm,ei,id,lc,li,np,ratio
    - dem_type:          Copernicus 30m Global DEM
    - date_strict:       True
    - etad:              False
    - etad_dir:          None
    - gdal_threads:      4
    - measurement:       gamma
    - ard_dir:           ARD
    - sar_dir:           SAR
    - tmp_dir:           TMP
    - wbm_dir:           WBM
    - logfile:           None
    METADATA
    - access_url:        None
    - doi:               None
    - licence:           None
    - processing_center: None
    """
    from s1ard.config import init as init_config
    extra = {ctx.args[i][2:]: ctx.args[i + 1]
             for i in range(0, len(ctx.args), 2)}
    init_config(target=config_file, source=config_source,
                overwrite=overwrite, **extra)


@cli.command(name='process',
             no_args_is_help=True,
             context_settings=dict(
                 ignore_unknown_options=True,
                 allow_extra_args=True, )
             )
@click.option('--config-file', '-c', required=False, type=click.Path(),
              help="Full path to an INI-style configuration text file. "
                   "If not defined, the package's default file will be used.")
@click.option('--debug', is_flag=True,
              help='Print debugging information for pyroSAR modules.')
@click.pass_context
def process(ctx, config_file, debug):
    """
    Central s1ard radar backscatter (rb) processing command.
    
    A config file can be defined from which all configuration is read.
    If not defined, configuration will be read from the package's default file.
    Additional options can be passed to override individual processing parameters
    in the configuration file. For example, to read all values from the configuration
    file except the acquisition mode and the annotation layers:
    
    s1rb process -c config.ini --acq_mode IW --annotation dm,id
    
    The snap_gpt_args argument can be provided by using a dedicated argument separator and quotes:
    
    s1rb process -c config.ini -- --snap_gpt_args "-J-Xmx100G -c 75G -q 30"
    
    \b
    The following defaults are set:
    PROCESSING
    - annotation:        dm,ei,id,lc,li,np,ratio
    - dem_type:          Copernicus 30m Global DEM
    - date_strict:       True
    - etad:              False
    - etad_dir:          None
    - gdal_threads:      4
    - snap_gpt_args:     None
    - measurement:       gamma
    - ard_dir:           ARD
    - sar_dir:           SAR
    - tmp_dir:           TMP
    - wbm_dir:           WBM
    - logfile:           None
    METADATA
    - access_url:        None
    - doi:               None
    - licence:           None
    - processing_center: None
    """
    from s1ard import process
    extra = {ctx.args[i][2:]: ctx.args[i + 1] for i in range(0, len(ctx.args), 2)}
    process(config_file=config_file, debug=debug, **extra)



import pytest
import requests
from s1ard.metadata.mapping import URL


def url_recursive(key, dictionary, parent_key=None):
    key_info = f"{parent_key}.{key}" if parent_key else key
    if isinstance(dictionary[key], dict):
        for k, v in dictionary[key].items():
            url_recursive(k, dictionary[key], key_info)
    else:
        response = requests.get(dictionary[key])
        assert response.status_code in [200, 418]


@pytest.mark.parametrize('url_key', URL.keys())
def test_url(url_key, dictionary=URL):
    url_recursive(url_key, dictionary)



from __future__ import annotations
from importlib.metadata import entry_points
from typing import Type
from .base import Processor

_GROUP = "s1ard.processors"


def available_processors() -> dict[str, str]:
    eps = entry_points().select(group=_GROUP)
    return {ep.name: f"{ep.module}:{ep.attr}" for ep in eps}


def load_processor(name: str) -> Type[Processor]:
    eps = {ep.name: ep for ep in entry_points().select(group=_GROUP)}
    if name not in eps:
        installed = ", ".join(sorted(eps)) or "(none)"
        raise ModuleNotFoundError(f"Processor '{name}' not found. Installed: {installed}")
    return eps[name].load()



import numpy as np
from osgeo import gdal, osr
from spatialist.ancillary import finder


def extract(src, dst, variable):
    """
    Extract an OCN product's image variable and write it to a new GeoTIFF file.
    Coordinates are extracted from the corresponding latitude and longitude
    image variables and the corner coordinates written as ground control
    points (GCPs) to the output file.

    Parameters
    ----------
    src: str
        path to OCN product SAFE folder
    dst: str
        the name of the GeoTIFF file to write
    variable: str
        name of the layer to extract from the OCN product, e.g. `owiNrcsCmod`

    Returns
    -------

    """
    ocn_target = finder(target=src, matchlist=['*.nc'])[0]
    ras_ocn = gdal.Open(f'NETCDF:{ocn_target}:{variable}')
    arr = ras_ocn.ReadAsArray()
    lines, samples = arr.shape
    
    driver = gdal.GetDriverByName('GTiff')
    ras_out = driver.CreateCopy(dst, ras_ocn)
    
    ras_lat = gdal.Open(f'NETCDF:{ocn_target}:{variable[:3]}Lat')
    ras_lon = gdal.Open(f'NETCDF:{ocn_target}:{variable[:3]}Lon')
    arr_lat = ras_lat.ReadAsArray()
    arr_lon = ras_lon.ReadAsArray()
    ras_lat = ras_lon = None
    xres = (np.max(arr_lon) - np.min(arr_lon)) / samples
    yres = (np.max(arr_lat) - np.min(arr_lat)) / lines
    # coordinates are at pixel center
    ulx = float(arr_lon[0, 0]) - xres / 2
    uly = float(arr_lat[0, 0]) + yres / 2
    urx = float(arr_lon[0, -1]) - xres / 2
    ury = float(arr_lat[0, -1]) + yres / 2
    llx = float(arr_lon[-1, 0]) - xres / 2
    lly = float(arr_lat[-1, 0]) + yres / 2
    lrx = float(arr_lon[-1, -1]) - xres / 2
    lry = float(arr_lat[-1, -1]) + yres / 2
    gcps = [gdal.GCP(ulx, uly, 0, 0, 0),
            gdal.GCP(urx, ury, 0, samples, 0),
            gdal.GCP(lrx, lry, 0, samples, lines),
            gdal.GCP(llx, lly, 0, 0, lines)]
    crs = osr.SpatialReference()
    crs.SetFromUserInput('EPSG:4326')
    ras_out.SetGCPs(gcps, crs)
    outband = ras_out.GetRasterBand(1)
    outband.SetNoDataValue(-999)
    outband.WriteArray(arr, 0, 0)
    del arr, arr_lat, arr_lon
    outband = None
    ras_out = None
    ras_ocn = None


def gapfill(src, dst, md, si):
    """
    Fill gaps of an image file using GDAL.

    Parameters
    ----------
    src: str
        the source image file
    dst: str
        the destination image file with gaps filled
    md: int
        the interpolation maximum distance
    si: int
        the number of smoothing iterations

    Returns
    -------

    See Also
    --------
    osgeo.gdal.FillNodata
    """
    ras = gdal.Open(src)
    band = ras.GetRasterBand(1)
    
    driver = gdal.GetDriverByName('GTiff')
    out = driver.CreateCopy(dst, ras)
    out_band = out.GetRasterBand(1)
    arr = band.ReadAsArray()
    out_band.WriteArray(arr)
    
    mask = band.GetMaskBand()
    result = gdal.FillNodata(out_band, mask, md, si)
    
    out_band = None
    out = None
    band = None
    ras = None



import os
import re
import shutil
from math import ceil
import copy
from lxml import etree
import numpy as np
from pyproj import Geod
from osgeo import gdal, gdalconst
from scipy.interpolate import griddata
from datetime import datetime
from dateutil.parser import parse as dateparse
from spatialist.envi import HDRobject
from spatialist.ancillary import finder
from pyroSAR import identify, identify_many
from pyroSAR.ancillary import Lock, LockCollection
from pyroSAR.snap.auxil import (gpt, parse_recipe,
                                parse_node, orb_parametrize)
from s1ard.config import keyval_check
from cesard.tile_extraction import aoi_from_scene
from cesard.ancillary import datamask
from cesard.snap import geo, gsr, mli, rtc, sgr, postprocess

# do not remove, needed for interface
from cesard.snap import find_datasets, lsm_encoding, version_dict

import logging

log = logging.getLogger('s1ard')


# main interface

def config_to_string(config):
    """
    Convert the values of a configuration dictionary to strings.

    Parameters
    ----------
    config: dict
        the configuration as returned by :func:`get_config_section`

    Returns
    -------
    dict
        the dictionary with the same structure but values converted to strings.
    """
    out = {}
    allowed = get_config_keys()
    for k, v in config.items():
        if k == 'dem_prepare_mode':
            continue
        if k not in allowed:
            raise ValueError(f"key '{k}' not in allowed keys: {allowed}")
        if v is None:
            out[k] = 'None'
        elif k in ['allow_res_osv', 'clean_edges',
                   'clean_edges_pixels', 'cleanup']:
            out[k] = str(v)
        elif k == 'gpt_args' and isinstance(v, list):
            out[k] = ' '.join(v)
        else:
            out[k] = v
    return out


def get_config_keys():
    """
    Get all allowed configuration keys.
    
    Returns
    -------
    List[str]
    """
    return ['allow_res_osv', 'clean_edges', 'clean_edges_pixels', 'cleanup',
            'dem_resampling_method', 'gpt_args', 'img_resampling_method']


def get_config_section(parser, **kwargs):
    """
    Get the`config.ini` `SNAP` section content as a dictionary.
    
    Parameters
    ----------
    parser: configparser.ConfigParser
    kwargs: dict[str]

    Returns
    -------
    dict
    """
    out = {}
    defaults = {
        'allow_res_osv': 'True',
        'cleanup': 'True',
        'clean_edges': 'True',
        'clean_edges_pixels': '4',
        'dem_resampling_method': 'BILINEAR_INTERPOLATION',
        'gpt_args': 'None',
        'img_resampling_method': 'BILINEAR_INTERPOLATION',
    }
    if 'SNAP' in parser.sections():
        section = parser['SNAP']
        for k, v in defaults.items():
            if k not in section and k not in kwargs:
                section[k] = v
    else:
        parser.read_dict({'SNAP': defaults})
        section = parser['SNAP']
    
    kwargs_str = config_to_string(kwargs)
    for k, v in kwargs_str.items():
        if k in get_config_keys():
            section[k] = v
    
    for k, v in section.items():
        v = keyval_check(key=k, val=v, allowed_keys=get_config_keys())
        if k == 'gpt_args':
            if v is not None:
                v = v.split(' ')
        if k == 'clean_edges_pixels':
            v = section.getint(k)
        if k in ['allow_res_osv', 'clean_edges', 'cleanup']:
            v = section.getboolean(k)
        out[k] = v
    out['dem_prepare_mode'] = 'single-4326'
    return out


def get_metadata(scene, outdir):
    """
    Get processing metadata needed for ARD product metadata.

    Parameters
    ----------
    scene: str
        the name of the SAR scene
    outdir: str
        the directory to search for processing output

    Returns
    -------
    dict
    """
    basename = os.path.splitext(os.path.basename(scene))[0]
    scenedir = os.path.join(outdir, basename)
    dim = finder(scenedir, ['*.dim'])[0]
    scene = identify(dim)
    rlks, azlks = scene.meta['looks']
    return {'azlks': azlks,
            'rlks': rlks}


def process(scene, outdir, measurement, spacing, dem,
            dem_resampling_method='BILINEAR_INTERPOLATION',
            img_resampling_method='BILINEAR_INTERPOLATION',
            rlks=None, azlks=None, tmpdir=None, export_extra=None,
            allow_res_osv=True, clean_edges=True, clean_edges_pixels=4,
            neighbors=None, gpt_args=None, cleanup=True):
    """
    Main function for SAR processing with SNAP.

    Parameters
    ----------
    scene: str
        The SAR scene file name.
    outdir: str
        The output directory for storing the final results.
    measurement: {'sigma', 'gamma'}
        the backscatter measurement convention:

        - gamma: RTC gamma nought (:math:`\\gamma^0_T`)
        - sigma: RTC sigma nought (:math:`\\sigma^0_T`)
    spacing: int or float
        The output pixel spacing in meters.
    dem: str
        The DEM filename. Can be created with :func:`s1ard.dem.mosaic`.
    dem_resampling_method: str
        The DEM resampling method.
    img_resampling_method: str
        The image resampling method.
    rlks: int or None
        The number of range looks.
    azlks: int or None
        The number of azimuth looks.
    tmpdir: str or None
        Path to a temporary directory for intermediate products.
    export_extra: list[str] or None
        A list of ancillary layers to create. Default None: do not create any ancillary layers.
        Options:

         - DEM
         - gammaSigmaRatio: :math:`\\sigma^0_T / \\gamma^0_T`
         - sigmaGammaRatio: :math:`\\gamma^0_T / \\sigma^0_T`
         - incidenceAngleFromEllipsoid
         - layoverShadowMask
         - localIncidenceAngle
         - NESZ: noise equivalent sigma zero
         - projectedLocalIncidenceAngle
         - scatteringArea
         - lookDirection: range look direction angle
    allow_res_osv: bool
        Also allow the less accurate RES orbit files to be used?
    clean_edges: bool
        Erode noisy image edges? See :func:`pyroSAR.snap.auxil.erode_edges`.
        Does not apply to layover-shadow mask.
    clean_edges_pixels: int
        The number of pixels to erode.
    neighbors: list[str] or None
        (only applies to GRD) an optional list of neighboring scenes to add
        a buffer around the main scene using function :func:`grd_buffer`.
        If GRDs are processed compeletely independently, gaps are introduced
        due to a missing overlap. If `neighbors` is None or an empty list,
        buffering is skipped.
    gpt_args: list[str] or None
        a list of additional arguments to be passed to the gpt call

        - e.g. ``['-x', '-c', '2048M']`` for increased tile cache size and intermediate clearing
    cleanup: bool
        Delete intermediate files after successful process termination?

    Returns
    -------

    Examples
    --------
    >>> from s1ard import snap
    >>> scene = 'S1A_IW_SLC__1SDV_20200103T170700_20200103T170727_030639_0382D5_6A12.zip'
    >>> dem = 'S1A_IW_SLC__1SDV_20200103T170700_20200103T170727_030639_0382D5_6A12_DEM_EEA10.tif'
    >>> outdir = '.'
    >>> spacing = 10
    >>> rlks = 5
    >>> azlks = 1
    >>> export_extra = ['localIncidenceAngle', 'incidenceAngleFromEllipsoid',
    >>>                 'scatteringArea', 'layoverShadowMask', 'gammaSigmaRatio']
    >>> snap.process(scene=scene, outdir=outdir, spacing=spacing, dem=dem,
    >>>              rlks=rlks, azlks=azlks, export_extra=export_extra)
    """
    if measurement not in ['gamma', 'sigma']:
        raise RuntimeError("'measurement' must either be 'gamma' or 'sigma'")
    if export_extra is None:
        export_extra = []
    basename = os.path.splitext(os.path.basename(scene))[0]
    outdir_scene = os.path.join(outdir, basename)
    if tmpdir is None:
        tmpdir = outdir
        tmpdir_scene = os.path.join(tmpdir, basename + '_tmp')
    else:
        tmpdir_scene = os.path.join(tmpdir, basename)
    os.makedirs(outdir_scene, exist_ok=True)
    os.makedirs(tmpdir_scene, exist_ok=True)
    
    out_base = os.path.join(outdir_scene, basename)
    tmp_base = os.path.join(tmpdir_scene, basename)
    
    id = identify(scene)
    workflows = []
    
    apply_rtc = True
    
    apply_grd_buffering = neighbors is not None and len(neighbors) > 0
    ############################################################################
    # general pre-processing
    out_pre = tmp_base + '_pre.dim'
    out_pre_wf = out_pre.replace('.dim', '.xml')
    workflows.append(out_pre_wf)
    output_noise = 'NESZ' in export_extra
    with Lock(out_pre):
        if not os.path.isfile(out_pre):
            log.info('preprocessing main scene')
            pre(src=scene, dst=out_pre, workflow=out_pre_wf,
                allow_res_osv=allow_res_osv, output_noise=output_noise,
                output_beta0=apply_rtc, gpt_args=gpt_args,
                add_slice_num=apply_grd_buffering)
        else:
            log.info('main scene has already been preprocessed')
    ############################################################################
    # GRD buffering
    if apply_grd_buffering:
        # general preprocessing of neighboring scenes
        out_pre_neighbors = []
        for item in neighbors:
            basename_nb = os.path.splitext(os.path.basename(item))[0]
            tmpdir_nb = os.path.join(tmpdir, basename_nb)
            os.makedirs(tmpdir_nb, exist_ok=True)
            tmp_base_nb = os.path.join(tmpdir_nb, basename_nb)
            out_pre_nb = tmp_base_nb + '_pre.dim'
            out_pre_nb_wf = out_pre_nb.replace('.dim', '.xml')
            with Lock(out_pre_nb):
                if not os.path.isfile(out_pre_nb):
                    log.info(f'preprocessing GRD neighbor: {item}')
                    pre(src=item, dst=out_pre_nb, workflow=out_pre_nb_wf,
                        allow_res_osv=allow_res_osv, output_noise=output_noise,
                        output_beta0=apply_rtc, gpt_args=gpt_args)
                else:
                    log.info(f'GRD neighbor has already been preprocessed: {item}')
            out_pre_neighbors.append(out_pre_nb)
        ########################################################################
        # buffering
        out_buffer = tmp_base + '_buf.dim'
        out_buffer_wf = out_buffer.replace('.dim', '.xml')
        with Lock(out_buffer):
            if not os.path.isfile(out_buffer):
                log.info('buffering GRD scene with neighboring acquisitions')
                with LockCollection(out_pre_neighbors + [out_pre], soft=True):
                    try:
                        grd_buffer(src=out_pre, dst=out_buffer, workflow=out_buffer_wf,
                                   neighbors=out_pre_neighbors, gpt_args=gpt_args,
                                   buffer=10 * spacing)
                        workflows.append(out_buffer_wf)
                        out_pre = out_buffer
                    except RuntimeError:
                        log.info('did not perform buffering because the slice number '
                                 'could not be determined')
                        pass
            else:
                out_pre = out_buffer
                log.info('GRD scene has already been buffered')
    ############################################################################
    # range look direction angle
    if 'lookDirection' in export_extra:
        with Lock(out_pre):
            look_direction(dim=out_pre)
    ############################################################################
    # multi-looking
    out_mli = tmp_base + '_mli.dim'
    out_mli_wf = out_mli.replace('.dim', '.xml')
    with Lock(out_pre, soft=True):
        with Lock(out_mli):
            if not os.path.isfile(out_mli):
                mli(src=out_pre, dst=out_mli, workflow=out_mli_wf,
                    spacing=spacing, rlks=rlks, azlks=azlks, gpt_args=gpt_args)
    if not os.path.isfile(out_mli):
        out_mli = out_pre
    else:
        workflows.append(out_mli_wf)
    ############################################################################
    # radiometric terrain flattening
    out_rtc = out_gsr = out_sgr = None
    if apply_rtc:
        out_rtc = tmp_base + '_rtc.dim'
        out_rtc_wf = out_rtc.replace('.dim', '.xml')
        workflows.append(out_rtc_wf)
        output_sigma0_rtc = measurement == 'sigma' or 'gammaSigmaRatio' in export_extra
        with LockCollection([out_mli, dem], soft=True):
            with Lock(out_rtc):
                if not os.path.isfile(out_rtc):
                    log.info('radiometric terrain correction')
                    rtc(src=out_mli, dst=out_rtc, workflow=out_rtc_wf, dem=dem,
                        dem_resampling_method=dem_resampling_method,
                        sigma0=output_sigma0_rtc,
                        scattering_area='scatteringArea' in export_extra,
                        gpt_args=gpt_args)
        ########################################################################
        # gamma-sigma ratio computation
        out_gsr = None
        if 'gammaSigmaRatio' in export_extra:
            out_gsr = tmp_base + '_gsr.dim'
            out_gsr_wf = out_gsr.replace('.dim', '.xml')
            workflows.append(out_gsr_wf)
            with Lock(out_rtc, soft=True):
                with Lock(out_gsr):
                    if not os.path.isfile(out_gsr):
                        log.info('computing gamma-sigma ratio')
                        gsr(src=out_rtc, dst=out_gsr, workflow=out_gsr_wf,
                            gpt_args=gpt_args)
        ########################################################################
        # sigma-gamma ratio computation
        out_sgr = None
        if 'sigmaGammaRatio' in export_extra:
            out_sgr = tmp_base + '_sgr.dim'
            out_sgr_wf = out_sgr.replace('.dim', '.xml')
            workflows.append(out_sgr_wf)
            with Lock(out_rtc, soft=True):
                with Lock(out_sgr):
                    if not os.path.isfile(out_sgr):
                        log.info('computing sigma-gamma ratio')
                        sgr(src=out_rtc, dst=out_sgr, workflow=out_sgr_wf,
                            gpt_args=gpt_args)
    ############################################################################
    # geocoding
    
    # Process to multiple UTM zones or just one?
    # For testing purposes only.
    utm_multi = True
    
    def run():
        out_geo = out_base + '_geo_{}.dim'.format(epsg)
        out_geo_wf = out_geo.replace('.dim', '.xml')
        sources = list(filter(None, [out_mli, out_rtc, out_gsr, out_sgr]))
        with LockCollection(sources, soft=True):
            with Lock(out_geo):
                if not os.path.isfile(out_geo):
                    log.info(f'geocoding to EPSG:{epsg}')
                    scene1 = identify(out_mli)
                    pols = scene1.polarizations
                    bands0 = ['NESZ_{}'.format(pol) for pol in pols]
                    if measurement == 'gamma':
                        bands1 = ['Gamma0_{}'.format(pol) for pol in pols]
                    else:
                        bands0.extend(['Sigma0_{}'.format(pol) for pol in pols])
                        bands1 = []
                    if 'scatteringArea' in export_extra:
                        bands1.append('simulatedImage')
                    if 'lookDirection' in export_extra:
                        bands0.append('lookDirection')
                    geo(*sources,
                        dst=out_geo, workflow=out_geo_wf,
                        spacing=spacing, crs=epsg, geometry=ext,
                        export_extra=export_extra,
                        standard_grid_origin_x=align_x,
                        standard_grid_origin_y=align_y,
                        bands0=bands0, bands1=bands1, dem=dem,
                        dem_resampling_method=dem_resampling_method,
                        img_resampling_method=img_resampling_method,
                        gpt_args=gpt_args)
                    log.info('edge cleaning')
                    postprocess(out_geo, clean_edges=clean_edges,
                                clean_edges_pixels=clean_edges_pixels)
                    log.info('creating valid data masks')
                    out_geo_data = out_geo.replace('.dim', '.data')
                    pattern = r'(?:Gamma0|Sigma0)_[VH]{2}\.img$'
                    measurements = finder(out_geo_data, [pattern],
                                          regex=True, recursive=False)
                    dm_ras = os.path.join(out_geo_data, 'datamask.tif')
                    dm_vec = dm_ras.replace('.tif', '.gpkg')
                    dm_vec = datamask(measurement=measurements[0],
                                      dm_ras=dm_ras, dm_vec=dm_vec)
                else:
                    log.info(f'geocoding to EPSG:{epsg} has already been performed')
        for wf in workflows:
            wf_dst = os.path.join(outdir_scene, os.path.basename(wf))
            if wf != wf_dst and not os.path.isfile(wf_dst):
                shutil.copyfile(src=wf, dst=wf_dst)
    
    log.info('determining UTM zone overlaps')
    aois = aoi_from_scene(scene=id, multi=utm_multi)
    for aoi in aois:
        ext = aoi['extent']
        epsg = aoi['epsg']
        align_x = aoi['extent_utm']['xmin']
        align_y = aoi['extent_utm']['ymax']
        run()
    ############################################################################
    # delete intermediate files
    if cleanup:
        log.info('cleaning up')
        if id.product == 'GRD':
            # delete everything except *_pre.* products which are reused for buffering
            # this needs to be improved so that these products are also removed if they
            # are no longer needed for any buffering.
            items = finder(target=tmpdir_scene, matchlist=['*'],
                           foldermode=1, recursive=False)
            for item in items:
                if not re.search(r'_pre\.', item):
                    if os.path.isfile(item):
                        os.remove(item)
                    else:
                        shutil.rmtree(item)
        else:
            shutil.rmtree(tmpdir_scene)


def translate_annotation(annotation, measurement):
    """
    Translate s1ard annotation keys to SAR processor naming.
    
    Parameters
    ----------
    annotation: List[str]
        the s1ard annotation keys (e.g. ei, gs)
    measurement: str
        the SAR backscatter measurement (gamma|sigma)

    Returns
    -------
    List[str]
        the annotation layer keys as required by the SAR processor
    """
    export_extra = None
    lookup = {'dm': 'layoverShadowMask',
              'ei': 'incidenceAngleFromEllipsoid',
              'lc': 'scatteringArea',
              'ld': 'lookDirection',
              'li': 'localIncidenceAngle',
              'np': 'NESZ',
              'gs': 'gammaSigmaRatio',
              'sg': 'sigmaGammaRatio'}
    
    if annotation is not None:
        annotation = [
            'gs' if x == 'ratio' and measurement == 'gamma'
            else 'sg' if x == 'ratio'
            else x for x in annotation]
        export_extra = []
        for layer in annotation:
            # supported by the SAR processor
            if layer in lookup:
                export_extra.append(lookup[layer])
            # supported by the NRB processor
            elif layer in ['em', 'id']:
                continue
            else:
                log.warning(f'unsupported annotation layer: {layer}')
    return export_extra


# processor-specific functions

def grd_buffer(src, dst, workflow, neighbors, buffer=100, gpt_args=None):
    """
    GRD extent buffering.
    GRDs, unlike SLCs, do not overlap in azimuth.
    With this function, a GRD can be buffered using the neighboring acquisitions.
    First, all images are mosaicked using the `SliceAssembly` operator
    and then subsetted to the extent of the main scene including a buffer.
    The `SliceAssembly` operator needs info about the slice number (i.e., the
    ID/position inside the data take). If the value in the metadata is 0 (which
    can be the case in NRT slicing mode), the slice number is determined using
    function :func:`~s1ard.snap.nrt_slice_num`. If this fails, the function will
    raise an error.

    Parameters
    ----------
    src: str
        the file name of the source scene in BEAM-DIMAP format.
    dst: str
        the file name of the target scene. Format is BEAM-DIMAP.
    workflow: str
        the output SNAP XML workflow filename.
    neighbors: list[str]
        the file names of neighboring scenes
    buffer: int
        the buffer size in meters
    gpt_args: list[str] or None
        a list of additional arguments to be passed to the gpt call

        - e.g. ``['-x', '-c', '2048M']`` for increased tile cache size
          and intermediate clearing

    Raises
    ------
    RuntimeError
        if the slice number of a scene is 0, and it could not be determined
        from the acquisition time

    Returns
    -------

    """
    if len(neighbors) == 0:
        raise RuntimeError("the list of 'neighbors' is empty")
    
    scenes = identify_many([src] + neighbors, sortkey='start')
    wf = parse_recipe('blank')
    ############################################
    read_ids = []
    for scene in scenes:
        read = parse_node('Read')
        read.parameters['file'] = scene.scene
        wf.insert_node(read)
        read_ids.append(read.id)
    ############################################
    asm = parse_node('SliceAssembly')
    wf.insert_node(asm, before=read_ids)
    ############################################
    id_main = [x.scene for x in scenes].index(src)
    buffer_px = int(ceil(buffer / scenes[0].spacing[1]))
    xmin = 0
    width = scenes[id_main].samples
    if id_main == 0:
        ymin = 0
        height = scenes[id_main].lines + buffer_px
    else:
        ymin = scenes[0].lines - buffer_px
        factor = 1 if id_main == len(scenes) - 1 else 2
        height = scenes[id_main].lines + buffer_px * factor
    sub = parse_node('Subset')
    sub.parameters['region'] = [xmin, ymin, width, height]
    sub.parameters['geoRegion'] = ''
    sub.parameters['copyMetadata'] = True
    wf.insert_node(sub, before=asm.id)
    ############################################
    write = parse_node('Write')
    wf.insert_node(write, before=sub.id)
    write.parameters['file'] = dst
    write.parameters['formatName'] = 'BEAM-DIMAP'
    ############################################
    wf.write(workflow)
    gpt(xmlfile=workflow, tmpdir=os.path.dirname(dst),
        gpt_args=gpt_args)


def look_direction(dim):
    """
    Compute the per-pixel range look direction angle.
    This adds a new layer to an existing BEAM-DIMAP product.

    Steps performed:

    - read geolocation grid points
    - limit grid point list to those relevant to the image
    - for each point, compute the range direction angle to the next point in range direction.
    - interpolate the grid to the full image dimensions

    Notes
    -----
    - The interpolation depends on the location of the grid points relative to the image.
      Hence, by subsetting the image by an amount of pixels/lines different to the grid point
      sampling rate, the first and last points will no longer be in the first and last line respectively.
    - The list might get very large when merging the scene with neighboring acquisitions using
      SliceAssembly and this longer list significantly changes the interpolation result.
      The difference in interpolation can be mitigated by reducing the list of points to
      those inside the image and those just outside of it.

    Parameters
    ----------
    dim: str
        a BEAM-DIMAP metadata file (extension .dim)

    Returns
    -------

    """
    
    def interpolate(infile, method='linear'):
        with open(infile, 'rb') as f:
            tree = etree.fromstring(f.read())
        
        lats = []
        lons = []
        lines = []
        pixels = []
        rgtimes = []
        aztimes = []
        
        pols = tree.xpath("//MDElem[@name='standAloneProductInformation']"
                          "/MDATTR[@name='transmitterReceiverPolarisation']")
        polarization = pols[0].text.lower()
        re_ns = "http://exslt.org/regular-expressions"
        ann_pol = tree.xpath(f"//MDElem[@name='annotation']"
                             f"//MDElem[re:test(@name, '-{polarization}-', 'i')]",
                             namespaces={'re': re_ns})
        nlines = int(tree.find('Raster_Dimensions/NROWS').text)
        npixels = int(tree.find('Raster_Dimensions/NCOLS').text)
        abstract = tree.xpath("//MDElem[@name='Abstracted_Metadata']")[0]
        
        for ann in ann_pol:
            points = ann.xpath(".//MDElem[@name='geolocationGridPoint']")
            for point in points:
                pixel = int(point.find("./MDATTR[@name='pixel']").text)
                line = int(point.find("./MDATTR[@name='line']").text)
                lat = float(point.find("./MDATTR[@name='latitude']").text)
                lon = float(point.find("./MDATTR[@name='longitude']").text)
                rgtime = float(point.find("./MDATTR[@name='slantRangeTime']").text)
                aztime = dateparse(point.find("./MDATTR[@name='azimuthTime']").text)
                aztime = (aztime - datetime(1900, 1, 1)).total_seconds()
                pixels.append(pixel)
                lines.append(line)
                rgtimes.append(rgtime)
                aztimes.append(aztime)
                lats.append(lat)
                lons.append(lon)
        coords = list(zip(rgtimes, aztimes))
        
        flt = abstract.xpath("./MDATTR[@name='first_line_time']")[0].text
        flt = (dateparse(flt) - datetime(1900, 1, 1)).total_seconds()
        llt = abstract.xpath("./MDATTR[@name='last_line_time']")[0].text
        llt = (dateparse(llt) - datetime(1900, 1, 1)).total_seconds()
        lti = float(abstract.xpath("./MDATTR[@name='line_time_interval']")[0].text)
        
        # limit the coords to those relevant to the image
        # (SliceAssembly extends the list but a subsequent Subset does not shorten it)
        az_before = [x[1] for x in coords if x[1] < flt]
        tmp_min = (max(az_before) if len(az_before) > 0 else flt) - lti
        az_after = [x[1] for x in coords if x[1] > llt]
        tmp_max = (min(az_after) if len(az_after) > 0 else llt) + lti
        
        coords_sub = [x for x in coords if tmp_min <= x[1] <= tmp_max]
        
        values = []
        coords_select = []
        g = Geod(ellps='WGS84')
        for i, v in enumerate(coords):
            if v in coords_sub:
                if i + 1 < len(coords) and pixels[i] < pixels[i + 1]:
                    az12, az21, dist = g.inv(lons[i], lats[i], lons[i + 1], lats[i + 1])
                    values.append(az12)
                else:
                    az12, az21, dist = g.inv(lons[i], lats[i], lons[i - 1], lats[i - 1])
                    values.append(az21)
                coords_select.append(v)
        
        coords = np.array(coords_select)
        values = np.array(values)
        
        rgtime_fl_max = max([v for i, v in enumerate(rgtimes) if lines[i] == 0])
        rgtime_ll_max = max([v for i, v in enumerate(rgtimes) if lines[i] == max(lines)])
        rgtime_max = min([rgtime_fl_max, rgtime_ll_max])
        
        xi = np.linspace(min(rgtimes), rgtime_max, npixels)
        yi = np.linspace(flt, llt, nlines)
        xi, yi = np.meshgrid(xi, yi)
        zi = griddata(coords, values, (xi, yi), method=method, fill_value=0)
        return zi
    
    def write(array, out, reference, format='ENVI'):
        src_dataset = gdal.Open(reference)
        cols = src_dataset.RasterXSize
        rows = src_dataset.RasterYSize
        driver = gdal.GetDriverByName(format)
        
        dst_dataset = driver.Create(out, cols, rows, 1, gdalconst.GDT_Float32)
        
        dst_dataset.SetMetadata(src_dataset.GetMetadata())
        dst_dataset.SetGeoTransform(src_dataset.GetGeoTransform())
        dst_dataset.SetProjection(src_dataset.GetProjection())
        if format == 'GTiff':
            dst_dataset.SetGCPs(src_dataset.GetGCPs(), src_dataset.GetGCPSpatialRef())
        else:
            array = array.astype('float32').byteswap().newbyteorder()
        dst_band = dst_dataset.GetRasterBand(1)
        dst_band.WriteArray(array)
        dst_band.FlushCache()
        dst_band = None
        src_dataset = None
        dst_dataset = None
        driver = None
        if format == 'ENVI':
            hdrfile = os.path.splitext(out)[0] + '.hdr'
            with HDRobject(hdrfile) as hdr:
                hdr.byte_order = 1
                hdr.description = 'Sentinel-1 EW Level-1 GRD Product - Unit: deg'
                hdr.band_names = 'lookDirection'
                hdr.write()
            auxfile = out + '.aux.xml'
            os.remove(auxfile)
    
    def metadata(dim):
        with open(dim, 'rb') as f:
            root = etree.fromstring(f.read())
        
        image_interp = root.find('Image_Interpretation')
        ld_search = image_interp.xpath('Spectral_Band_Info[BAND_NAME="lookDirection"]')
        
        if len(ld_search) == 0:
            # number of bands
            element_bands = root.find('Raster_Dimensions/NBANDS')
            bands = int(element_bands.text)
            element_bands.text = str(bands + 1)
            
            # data access
            data_access = root.find('Data_Access')
            data_files = data_access.findall('./Data_File')
            last = data_files[-1]
            new = copy.deepcopy(last)
            fpath = new.find('DATA_FILE_PATH').attrib['href']
            fpath = fpath.replace(os.path.basename(fpath), 'lookDirection.hdr')
            new.find('DATA_FILE_PATH').attrib['href'] = fpath
            new.find('BAND_INDEX').text = str(bands)
            data_access.insert(data_access.index(last) + 1, new)
            
            # band info
            info = etree.SubElement(image_interp, 'Spectral_Band_Info')
            npixels = root.find('Raster_Dimensions/NCOLS').text
            nlines = root.find('Raster_Dimensions/NROWS').text
            etree.SubElement(info, 'BAND_INDEX').text = str(bands)
            etree.SubElement(info, 'BAND_DESCRIPTION').text = 'range look direction'
            etree.SubElement(info, 'BAND_NAME').text = 'lookDirection'
            etree.SubElement(info, 'BAND_RASTER_WIDTH').text = npixels
            etree.SubElement(info, 'BAND_RASTER_HEIGHT').text = nlines
            etree.SubElement(info, 'DATA_TYPE').text = 'float32'
            etree.SubElement(info, 'PHYSICAL_UNIT').text = 'deg'
            etree.SubElement(info, 'LOG10_SCALED').text = 'false'
            etree.SubElement(info, 'NO_DATA_VALUE_USED').text = 'true'
            etree.SubElement(info, 'NO_DATA_VALUE').text = '0.0'
            
            etree.indent(root, space='    ')
            tree = etree.ElementTree(root)
            tree.write(dim, pretty_print=True, xml_declaration=True, encoding='utf-8')
    
    data = dim.replace('.dim', '.data')
    out = os.path.join(data, 'lookDirection.img')
    if not os.path.isfile(out):
        log.info('look direction computation')
        ref = finder(target=data, matchlist=['*.img'])[0]
        arr = interpolate(infile=dim)
        write(array=arr, out=out, reference=ref)
        metadata(dim=dim)
    else:
        log.info('look direction has already been computed')


def nrt_slice_num(dim):
    """
    Check whether a product has a non-zero slice number and add it if not.
    In NRT Slicing mode, both `sliceNumber` and `totalSlices` are 0 in the manifest.safe file.
    `sliceNumber` is however needed in function :func:`~s1ard.snap.grd_buffer` for
    the SNAP operator `SliceAssembly`.
    The time from `segmentStartTime` to `last_line_time` is divided by
    the acquisition duration (`last_line_time` - `first_line_time`).
    `totalSlices` is set to 100, which is expected to exceed the maximum possible value.

    Parameters
    ----------
    dim: str
        the scene in BEAM-DIMAP format

    Raises
    ------
    RuntimeError
        if the slice number is 0, and it cannot be computed because
        the segment start time cannot be read from the metadata

    Returns
    -------

    """
    with open(dim, 'rb') as f:
        root = etree.fromstring(f.read())
    abstract = root.xpath("//MDElem[@name='Abstracted_Metadata']")[0]
    slice_num = abstract.xpath("./MDATTR[@name='slice_num']")[0]
    if slice_num.text == '0':
        flt = dateparse(abstract.xpath("./MDATTR[@name='first_line_time']")[0].text)
        llt = dateparse(abstract.xpath("./MDATTR[@name='last_line_time']")[0].text)
        try:
            sst = dateparse(root.xpath("//MDATTR[@name='segmentStartTime']")[0].text)
        except IndexError:
            raise RuntimeError('could not determine slice number '
                               'due to missing segment start time')
        aqd = llt - flt
        slice_num_new = str(int(round((llt - sst) / aqd)))
        slice_num.text = slice_num_new
        for item in root.xpath("//MDATTR[@name='sliceNumber']"):
            item.text = slice_num_new
        for item in root.xpath("//MDATTR[@name='totalSlices']"):
            item.text = '100'
        etree.indent(root, space='    ')
        tree = etree.ElementTree(root)
        tree.write(dim, pretty_print=True, xml_declaration=True,
                   encoding='utf-8')


def pre(src, dst, workflow, allow_res_osv=True, osv_continue_on_fail=False,
        output_noise=True, output_beta0=True, output_sigma0=True,
        output_gamma0=False, add_slice_num=True, gpt_args=None):
    """
    General SAR preprocessing. The following operators are used (optional steps in brackets):
    Apply-Orbit-File(->Remove-GRD-Border-Noise)->Calibration->ThermalNoiseRemoval(->TOPSAR-Deburst)

    Parameters
    ----------
    src: str
        the file name of the source scene
    dst: str
        the file name of the target scene. Format is BEAM-DIMAP.
    workflow: str
        the output SNAP XML workflow filename.
    allow_res_osv: bool
        Also allow the less accurate RES orbit files to be used?
    osv_continue_on_fail: bool
        Continue processing if no OSV file can be downloaded or raise an error?
    output_noise: bool
        output the noise power images?
    output_beta0: bool
        output beta nought backscatter needed for RTC?
    output_sigma0: bool
        output sigma nought backscatter needed for NESZ?
    output_gamma0: bool
        output gamma nought backscatter needed?
    add_slice_num: bool
        determine a slice number and add it to the product's metadata?
        This is only necessary if GRD buffering is intended.
        See :func:`~s1ard.snap.nrt_slice_num`.
    gpt_args: list[str] or None
        a list of additional arguments to be passed to the gpt call
        
        - e.g. ``['-x', '-c', '2048M']`` for increased tile cache size and intermediate clearing
    
    Returns
    -------

    See Also
    --------
    pyroSAR.snap.auxil.orb_parametrize
    """
    scene = identify(src)
    if not os.path.isfile(workflow):
        polarizations = scene.polarizations
        wf = parse_recipe('blank')
        ############################################
        read = parse_node('Read')
        read.parameters['file'] = scene.scene
        wf.insert_node(read)
        ############################################
        orb = orb_parametrize(scene=scene, formatName='SENTINEL-1',
                              allow_RES_OSV=allow_res_osv,
                              continueOnFail=osv_continue_on_fail)
        wf.insert_node(orb, before=read.id)
        last = orb
        ############################################
        if re.search('S1[A-Z]', scene.sensor) and scene.product == 'GRD':
            bn = parse_node('Remove-GRD-Border-Noise')
            wf.insert_node(bn, before=last.id)
            bn.parameters['selectedPolarisations'] = polarizations
            last = bn
        ############################################
        cal = parse_node('Calibration')
        wf.insert_node(cal, before=last.id)
        cal.parameters['selectedPolarisations'] = polarizations
        cal.parameters['outputBetaBand'] = output_beta0
        cal.parameters['outputSigmaBand'] = output_sigma0
        cal.parameters['outputGammaBand'] = output_gamma0
        ############################################
        tnr = parse_node('ThermalNoiseRemoval')
        wf.insert_node(tnr, before=cal.id)
        tnr.parameters['outputNoise'] = output_noise
        last = tnr
        ############################################
        if scene.product == 'SLC' and scene.acquisition_mode in ['EW', 'IW']:
            deb = parse_node('TOPSAR-Deburst')
            wf.insert_node(deb, before=last.id)
            last = deb
        ############################################
        write = parse_node('Write')
        wf.insert_node(write, before=last.id)
        write.parameters['file'] = dst
        write.parameters['formatName'] = 'BEAM-DIMAP'
        ############################################
        wf.write(workflow)
    if not os.path.isfile(dst):
        gpt(xmlfile=workflow, tmpdir=os.path.dirname(dst),
            gpt_args=gpt_args, removeS1BorderNoiseMethod='ESA')
        if scene.product == 'GRD' and add_slice_num:
            try:
                nrt_slice_num(dim=dst)
            except RuntimeError:
                raise RuntimeError('cannot obtain slice number')



import os
import re
from pathlib import Path
from dateutil.parser import parse as dateparse
from packaging.version import Version
from pystac_client import Client
from pystac_client.stac_api_io import StacApiIO
from spatialist.vector import Vector
from shapely.geometry import shape
from pyroSAR import identify_many
from cesard.ancillary import date_to_utc, buffer_time
from cesard.search import asf_select
import logging

log = logging.getLogger('s1ard')


class STACArchive(object):
    """
    Search for scenes in a SpatioTemporal Asset Catalog.
    Scenes are expected to be unpacked with a folder suffix .SAFE.
    The interface is kept consistent with :class:`~s1ard.search.ASFArchive`,
    :class:`~s1ard.search.STACParquetArchive` and :class:`pyroSAR.drivers.Archive`.
    
    Parameters
    ----------
    url: str
        the catalog URL
    collections: str or list[str]
        the catalog collection(s) to be searched
    timeout: int
        the allowed timeout in seconds
    max_retries: int or None
        the number of times to retry requests. Set to None to disable retries.
    
    See Also
    --------
    pystac_client.Client.open
    pystac_client.stac_api_io.StacApiIO
    """
    
    def __init__(self, url, collections, timeout=60, max_retries=20):
        self.url = url
        self.timeout = timeout
        self.max_tries = max_retries
        self._open_catalog()
        if isinstance(collections, str):
            self.collections = [collections]
        elif isinstance(collections, list):
            self.collections = collections
        else:
            raise TypeError("'collections' must be of type str or list")
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
    
    def _filter_duplicates(self, values):
        tmp = sorted(values, key=lambda x: os.path.basename(x[-2]))
        pattern = '([0-9A-Z_]{16})_([0-9T]{15})_([0-9T]{15})'
        keep = []
        i = 0
        while i < len(tmp):
            group = [tmp[i]]
            base = os.path.basename(tmp[i][-2])
            match1 = re.search(pattern, base).groups()
            j = i + 1
            while j < len(tmp):
                base = os.path.basename(tmp[j][-2])
                match2 = re.search(pattern, base).groups()
                if match1 == match2:
                    group.append(tmp[j])
                    j += 1
                else:
                    break
            if len(group) > 1:
                tproc = [x[-1] for x in group]
                keep.append(group[tproc.index(max(tproc))])
            else:
                keep.append(group[0])
            i = j
        return keep
    
    def _open_catalog(self):
        stac_api_io = StacApiIO(max_retries=self.max_tries)
        self.catalog = Client.open(url=self.url,
                                   stac_io=stac_api_io,
                                   timeout=self.timeout)
    
    def close(self):
        del self.catalog
    
    def select(self, sensor=None, product=None, acquisition_mode=None,
               mindate=None, maxdate=None, frameNumber=None,
               vectorobject=None, date_strict=True, check_exist=True,
               return_value="scene"):
        """
        Select scenes from the catalog. Duplicates (same acquisition time) are filtered
        by returning only the last processed product. Used STAC property keys:

        - platform
        - start_datetime
        - end_datetime
        - created
        - sar:instrument_mode
        - sar:product_type
        - s1:datatake (custom)

        Parameters
        ----------
        sensor: str or list[str] or None
            S1A or S1B
        product: str or list[str] or None
            GRD or SLC
        acquisition_mode: str or list[str] or None
            IW, EW or SM
        mindate: str or datetime.datetime or None
            the minimum acquisition date; timezone-unaware dates are interpreted as UTC.
        maxdate: str or datetime.datetime or None
            the maximum acquisition date; timezone-unaware dates are interpreted as UTC.
        frameNumber: int or str or list[int or str] or None
            the data take ID in decimal (int) or hexadecimal (str) representation.
            Requires custom STAC key `s1:datatake`.
        vectorobject: spatialist.vector.Vector or None
            a geometry with which the scenes need to overlap. The object may only contain one feature.
        date_strict: bool
            treat dates as strict limits or also allow flexible limits to incorporate scenes
            whose acquisition period overlaps with the defined limit?

            - strict: start >= mindate & stop <= maxdate
            - not strict: stop >= mindate & start <= maxdate
        check_exist: bool
            check whether found files exist locally?
        return_value: str or List[str]
            the query return value(s). Options:
            
            - acquisition_mode: the sensor's acquisition mode, e.g., IW, EW, SM
            - frameNumber: the frame or datatake number
            - geometry_wkb: the scene's footprint geometry formatted as WKB
            - geometry_wkt: the scene's footprint geometry formatted as WKT
            - mindate: the acquisition start datetime in UTC formatted as YYYYmmddTHHMMSS
            - maxdate: the acquisition end datetime in UTC formatted as YYYYmmddTHHMMSS
            - product: the product type, e.g., SLC, GRD
            - scene: the scene's storage location path (default)
            - sensor: the satellite platform, e.g., S1A or S1B

        Returns
        -------
        list or list[tuple]
            If a single return_value is specified: list of values
            If multiple return_values are specified: list of tuples containing the requested values

        See Also
        --------
        pystac_client.Client.search
        """
        pars = locals()
        del pars['date_strict']
        del pars['check_exist']
        del pars['return_value']
        del pars['self']
        
        if isinstance(return_value, str):
            return_values = [return_value]
        else:
            return_values = return_value
        
        lookup = {'product': 'sar:product_type',
                  'acquisition_mode': 'sar:instrument_mode',
                  'mindate': 'start_datetime',
                  'maxdate': 'end_datetime',
                  'sensor': 'platform',
                  'frameNumber': 's1:datatake'}
        lookup_platform = {
            'S1A': 'sentinel-1a',
            'S1B': 'sentinel-1b',
            'S1C': 'sentinel-1c',
            'S1D': 'sentinel-1Dd'
        }
        lookup_platform_reverse = {value: key for key, value
                                   in lookup_platform.items()}
        
        args = {'datetime': [None, None]}
        flt = {'op': 'and', 'args': []}
        for key in pars.keys():
            val = pars[key]
            if val is None:
                continue
            if key in ['mindate', 'maxdate']:
                val = date_to_utc(val, str_format='%Y-%m-%dT%H:%M:%SZ')
            if key == 'mindate':
                args['datetime'][0] = val
                if date_strict:
                    arg = {'op': '>=', 'args': [{'property': 'start_datetime'}, val]}
                else:
                    arg = {'op': '>=', 'args': [{'property': 'end_datetime'}, val]}
                flt['args'].append(arg)
            elif key == 'maxdate':
                args['datetime'][1] = val
                if date_strict:
                    arg = {'op': '<=', 'args': [{'property': 'end_datetime'}, val]}
                else:
                    arg = {'op': '<=', 'args': [{'property': 'start_datetime'}, val]}
                flt['args'].append(arg)
            elif key == 'vectorobject':
                if isinstance(val, Vector):
                    if val.nfeatures > 1:
                        raise RuntimeError("'vectorobject' may only contain one feature")
                    with val.clone() as vec:
                        vec.reproject(4326)
                        feat = vec.getFeatureByIndex(0)
                        json = feat.ExportToJson(as_object=True)['geometry']
                        feat = None
                        args['intersects'] = json
                else:
                    raise TypeError('argument vectorobject must be of type spatialist.vector.Vector')
            else:
                if isinstance(val, (str, int)):
                    val = [val]
                val_format = []
                for v in val:
                    if key == 'sensor':
                        v = lookup_platform[v]
                    if key == 'frameNumber' and isinstance(v, int):
                        v = '{:06X}'.format(v)  # convert to hexadecimal
                    val_format.append(v)
                if len(val_format) == 1:
                    arg = {'op': '=', 'args': [{'property': lookup[key]}, val_format[0]]}
                else:
                    arg = {'op': 'in', 'args': [{'property': lookup[key]}, val_format]}
                flt['args'].append(arg)
        if len(flt['args']) == 0:
            flt = None
        if args['datetime'] == [None, None]:
            args['datetime'] = None
        result = self.catalog.search(collections=self.collections,
                                     filter=flt, max_items=None,
                                     **args)
        result = list(result.items())
        out = []
        for item in result:
            assets = item.assets
            ref = assets[list(assets.keys())[0]]
            href = ref.href
            scene = href[:re.search(r'\.SAFE', href).end()]
            scene = re.sub('^file://', '', scene)
            if Path(scene).exists():
                scene = os.path.realpath(scene)
            else:
                if check_exist:
                    raise RuntimeError('scene does not exist locally:', scene)
            
            # prepare return values
            values = []
            for key in return_values:
                if key == "scene":
                    values.append(scene)
                elif key in lookup.keys():
                    value = item.properties[lookup[key]]
                    if key in ['mindate', 'maxdate']:
                        value = dateparse(value).strftime('%Y%m%dT%H%M%S')
                    if key == 'sensor':
                        value = lookup_platform_reverse[value]
                    values.append(value)
                # reverse coordinate order to be consistent with ASF return
                elif key == "geometry_wkt":
                    values.append(shape(item.geometry).wkt)
                elif key == "geometry_wkb":
                    values.append(shape(item.geometry).wkb)
                else:
                    raise ValueError(f"Invalid return value: {key}")
            
            values.append(scene)
            created = dateparse(item.properties['created'])
            values.append(created)
            
            out.append(values)
        
        out = self._filter_duplicates(out)
        
        def reduce(i):
            o = i[:-2]
            return o[0] if len(o) == 1 else tuple(o)
        
        out = [reduce(x) for x in out]
        return out


class STACParquetArchive(object):
    """
    Search for scenes in STAC geoparquet dump.
    Scenes are expected to be unpacked with a folder suffix .SAFE.
    The interface is kept consistent with :class:`~s1ard.search.ASFArchive`,
    :class:`~s1ard.search.STACArchive` and :class:`pyroSAR.drivers.Archive`.

    Parameters
    ----------
    files: str
        the file search pattern, e.g. `/path/to/*parquet`
    """
    
    def __init__(self, files):
        self.files = files
    
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        return
    
    def close(self):
        pass
    
    def select(self, sensor=None, product=None, acquisition_mode=None,
               mindate=None, maxdate=None, frameNumber=None,
               vectorobject=None, date_strict=True, return_value='scene'):
        """
        Select scenes from a STAC catalog's geoparquet dump. Used STAC property keys:

        - platform
        - start_datetime
        - end_datetime
        - sar:instrument_mode
        - sar:product_type
        - s1:datatake (custom)

        Parameters
        ----------
        sensor: str or list[str] or None
            S1A or S1B
        product: str or list[str] or None
            GRD or SLC
        acquisition_mode: str or list[str] or None
            IW, EW or SM
        mindate: str or datetime.datetime or None
            the minimum acquisition date; timezone-unaware dates are interpreted as UTC.
        maxdate: str or datetime.datetime or None
            the maximum acquisition date; timezone-unaware dates are interpreted as UTC.
        frameNumber: int or str or list[int or str] or None
            the data take ID in decimal (int) or hexadecimal (str) representation.
            Requires custom STAC key `s1:datatake`.
        vectorobject: spatialist.vector.Vector or None
            a geometry with which the scenes need to overlap
        date_strict: bool
            treat dates as strict limits or also allow flexible limits to incorporate scenes
            whose acquisition period overlaps with the defined limit?

            - strict: start >= mindate & stop <= maxdate
            - not strict: stop >= mindate & start <= maxdate
        return_value: str or List[str]
            the query return value(s). Options:
            
            - acquisition_mode: the sensor's acquisition mode, e.g., IW, EW, SM
            - frameNumber: the frame or datatake number
            - geometry_wkb: the scene's footprint geometry formatted as WKB
            - geometry_wkt: the scene's footprint geometry formatted as WKT
            - mindate: the acquisition start datetime in UTC formatted as YYYYmmddTHHMMSS
            - maxdate: the acquisition end datetime in UTC formatted as YYYYmmddTHHMMSS
            - product: the product type, e.g., SLC, GRD
            - scene: the scene's storage location path (default)
            - sensor: the satellite platform, e.g., S1A or S1B

        Returns
        -------
        List[str] or List[tuple[str]]
            the selected return value(s). Depending on whether a single or multiple
            values have been defined for `return_value`, the returned list will
            contain strings or tuples.
        
        See Also
        --------
        stac_geoparquet.arrow.to_parquet
        duckdb.query
        """
        pars = locals()
        try:
            import duckdb
        except ImportError:
            raise ImportError("this method requires 'duckdb>=1.1.1' to be installed")
        ddb_version = Version(duckdb.__version__)
        ddb_version_req = Version('1.1.1')
        if ddb_version < ddb_version_req:
            raise ImportError("duckdb version must be >= 1.1.1")
        
        duckdb.install_extension('spatial')
        duckdb.load_extension('spatial')
        
        del pars['self']
        del pars['date_strict']
        del pars['return_value']
        
        lookup = {'product': 'sar:product_type',
                  'acquisition_mode': 'sar:instrument_mode',
                  'mindate': 'start_datetime',
                  'maxdate': 'end_datetime',
                  'sensor': 'platform',
                  'frameNumber': 's1:datatake'}
        lookup_platform = {'S1A': 'sentinel-1a',
                           'S1B': 'sentinel-1b',
                           'S1C': 'sentinel-1c',
                           'S1D': 'sentinel-1d'}
        return_value_mapping = {
            "geometry_wkb": "ST_AsWKB(geometry)",
            "geometry_wkt": "ST_AsText(geometry)",
            "mindate": f"STRFTIME({lookup['mindate']} AT TIME ZONE 'UTC', '%Y%m%dT%H%M%S')",
            "maxdate": f"STRFTIME({lookup['maxdate']} AT TIME ZONE 'UTC', '%Y%m%dT%H%M%S')",
            "scene": "replace(json_extract_string(assets::json, '$.folder.href'), 'file://', '')"
        }
        for k, v in lookup.items():
            if k not in return_value_mapping:
                return_value_mapping[k] = f"\"{lookup[k]}\""
        
        return_values = return_value if isinstance(return_value, list) else [return_value]
        for return_value in return_values:
            if return_value not in return_value_mapping:
                raise ValueError(f"unsupported return value '{return_value}'.\n"
                                 f"supported options: {list(return_value_mapping.keys())}")
        
        terms = []
        for key in pars.keys():
            val = pars[key]
            if val is None:
                continue
            if key in ['mindate', 'maxdate']:
                val = date_to_utc(val, str_format='%Y-%m-%dT%H:%M:%SZ')
            if key == 'mindate':
                if date_strict:
                    terms.append(f'"start_datetime" >= \'{val}\'')
                else:
                    terms.append(f'"end_datetime" >= \'{val}\'')
            elif key == 'maxdate':
                if date_strict:
                    terms.append(f'"end_datetime" <= \'{val}\'')
                else:
                    terms.append(f'"start_datetime" <= \'{val}\'')
            elif key == 'vectorobject':
                if isinstance(val, Vector):
                    if val.nfeatures > 1:
                        raise RuntimeError("'vectorobject' may only contain one feature")
                    with val.clone() as tmp:
                        tmp.reproject(4326)
                        wkt = tmp.convert2wkt(set3D=False)[0]
                    terms.append(f'ST_Intersects(geometry, '
                                 f'ST_GeomFromText(\'{wkt}\'))')
                else:
                    raise TypeError('argument vectorobject must be of type spatialist.vector.Vector')
            else:
                if isinstance(val, (str, int)):
                    val = [val]
                val_format = []
                for v in val:
                    if key == 'sensor':
                        v = lookup_platform[v]
                    if key == 'frameNumber' and isinstance(v, int):
                        v = '{:06X}'.format(v)  # convert to hexadecimal
                    val_format.append(v)
                if len(val_format) == 1:
                    subterm = f'"{lookup[key]}" = \'{val_format[0]}\''
                else:
                    subterm = f'"{lookup[key]}" IN {tuple(val_format)}'
                terms.append(subterm)
        sql_where = ' AND '.join(terms)
        sql_return_value = ', '.join([return_value_mapping[x] for x in return_values])
        sql_query = f"""
        SELECT {sql_return_value}
        FROM '{self.files}' WHERE {sql_where}
        """
        result = duckdb.query(sql_query).fetchall()
        if 'sensor' in return_values:
            lookup_platform_reverse = {value: key for key, value in lookup_platform.items()}
            sensor_index = return_values.index('sensor')
            for i, item in enumerate(result):
                item_new = list(item)
                item_new[sensor_index] = lookup_platform_reverse[item[sensor_index]]
                result[i] = tuple(item_new)
        if len(return_values) == 1:
            out = [x[0] for x in result]
        else:
            out = result
        return sorted(out)


def collect_neighbors(archive, scene, stac_check_exist=True):
    """
    Collect a scene's neighboring acquisitions in a data take.
    
    Parameters
    ----------
    archive: pyroSAR.drivers.Archive or STACArchive or STACParquetArchive or ASFArchive
        an open scene archive connection
    scene: pyroSAR.drivers.ID
        the Sentinel-1 scene to be checked
    stac_check_exist: bool
        if `archive` is of type :class:`STACArchive`, check the local existence of the scenes?

    Returns
    -------
    list[str]
        the filenames/URLs of the neighboring scenes
    """
    start, stop = buffer_time(scene.start, scene.stop, seconds=2)
    
    kwargs = {'mindate': start, 'maxdate': stop, 'date_strict': False,
              'sensor': scene.sensor, 'product': scene.product,
              'acquisition_mode': scene.acquisition_mode,
              'return_value': ['scene', 'mindate', 'maxdate']}
    if isinstance(archive, STACArchive):
        kwargs['check_exist'] = stac_check_exist
    
    selection = archive.select(**kwargs)
    neighbors = [x for x in selection if x[1] != scene.start]
    if len(neighbors) > 2:
        # more than two neighbors can exist if multiple versions of the
        # datatake with different slicing exist.
        start_ref = dateparse(scene.start)
        stop_ref = dateparse(scene.stop)
        start_diff = []
        stop_diff = []
        for neighbor, start, stop in neighbors:
            start_diff.append(abs(start_ref - dateparse(stop)))
            stop_diff.append(abs(stop_ref - dateparse(start)))
        predecessor = neighbors[start_diff.index(min(start_diff))]
        successor = neighbors[stop_diff.index(min(stop_diff))]
        neighbors = [predecessor, successor]
    return [x[0] for x in neighbors]


def check_acquisition_completeness(archive, scenes):
    """
    Check presence of neighboring acquisitions.
    Check that for each scene a predecessor and successor can be queried
    from the database unless the scene is at the start or end of the data take.
    This ensures that no scene that could be covering an area of interest is missed
    during processing. In case a scene is suspected to be missing, the Alaska Satellite Facility (ASF)
    online catalog is cross-checked.
    An error will only be raised if the locally missing scene is present in the ASF catalog.
    It may happen that a neighbor is missing and this error is not raised if the scene is also
    missing on ASF.

    Parameters
    ----------
    archive: pyroSAR.drivers.Archive or STACArchive
        an open scene archive connection
    scenes: list[pyroSAR.drivers.ID]
        a list of scenes

    Returns
    -------

    Raises
    ------
    RuntimeError

    See Also
    --------
    s1ard.search.asf_select
    """
    messages = []
    for scene in scenes:
        log.debug(f'checking acquisition completeness for scene {scene.scene}')
        slice = scene.meta['sliceNumber']
        n_slices = scene.meta['totalSlices']
        groupsize = 3
        has_successor = True
        has_predecessor = True
        
        start, stop = buffer_time(scene.start, scene.stop, seconds=2)
        ref = None
        if slice == 0 or n_slices == 0:
            # NRT slicing mode
            ref = asf_select(sensor=scene.sensor,
                             product=scene.product,
                             acquisition_mode=scene.acquisition_mode,
                             mindate=start,
                             maxdate=stop,
                             return_value='scene')
            if len(ref) > 0:
                ref = [os.path.basename(x).replace('.zip', '.SAFE') for x in ref]
                match = [re.search(scene.pattern, x).groupdict() for x in ref]
                ref_start_min = min([x['start'] for x in match])
                ref_stop_max = max([x['stop'] for x in match])
                if ref_start_min == scene.start:
                    groupsize -= 1
                    has_predecessor = False
                if ref_stop_max == scene.stop:
                    groupsize -= 1
                    has_successor = False
            else:
                # don't assume neighbors if no products could be found of ASF
                has_successor = has_predecessor = False
        else:
            if slice == 1:  # first slice in the data take
                groupsize -= 1
                has_predecessor = False
            if slice == n_slices:  # last slice in the data take
                groupsize -= 1
                has_successor = False
        # Do another database selection to get the scene in question as well as its potential
        # predecessor and successor by adding an acquisition time buffer of two seconds.
        group = archive.select(sensor=scene.sensor,
                               product=scene.product,
                               acquisition_mode=scene.acquisition_mode,
                               mindate=start,
                               maxdate=stop,
                               date_strict=False)
        group = identify_many(group)
        # if the number of selected scenes is lower than the expected group size,
        # check whether the predecessor, the successor or both are missing by
        # cross-checking with the ASF database.
        if len(group) < groupsize:
            if ref is None:
                ref = asf_select(sensor=scene.sensor,
                                 product=scene.product,
                                 acquisition_mode=scene.acquisition_mode,
                                 mindate=start,
                                 maxdate=stop,
                                 return_value='scene')
            if len(ref) > 0:
                ref = [os.path.basename(x).replace('.zip', '.SAFE') for x in ref]
                match = [re.search(scene.pattern, x).groupdict() for x in ref]
                ref_start_min = min([x['start'] for x in match])
                ref_stop_max = max([x['stop'] for x in match])
                start_min = min([x.start for x in group])
                stop_max = max([x.stop for x in group])
                missing = []
                if ref_start_min < start < start_min and has_predecessor:
                    missing.append('predecessor')
                if stop_max < stop < ref_stop_max and has_successor:
                    missing.append('successor')
                if len(missing) > 0:
                    base = os.path.basename(scene.scene)
                    messages.append(f'{" and ".join(missing)} acquisition for scene {base}')
    if len(messages) != 0:
        text = '\n - '.join(messages)
        raise RuntimeError(f'missing the following scenes:\n - {text}')



import os
import re
import copy
import importlib.resources
from datetime import datetime, timedelta
import configparser
from dateutil.parser import parse as dateparse
from osgeo import gdal
from s1ard.processors.registry import load_processor
from cesard.config import keyval_check, validate_options, validate_value
from typing import Any


def _get_config_metadata(parser, **kwargs):
    # METADATA section
    allowed_keys = get_keys(section='metadata')
    if 'METADATA' not in parser.sections():
        parser.add_section('METADATA')
    meta_sec = parser['METADATA']
    
    # override config file parameters
    for k, v in kwargs.items():
        if k in allowed_keys:
            meta_sec[k] = v.strip()
    
    # set defaults
    if 'format' not in meta_sec.keys():
        meta_sec['format'] = 'OGC, STAC'
    if 'copy_original' not in meta_sec.keys():
        meta_sec['copy_original'] = 'True'
    
    out = {}
    for k, v in meta_sec.items():
        v = keyval_check(key=k, val=v, allowed_keys=allowed_keys)
        if k == 'format':
            v = meta_sec.get_list(k)
        if k == 'copy_original':
            v = meta_sec.getboolean(k)
        out[k] = v
    for key in allowed_keys:
        if key not in out.keys():
            out[key] = None
    return out


def _get_config_processing(parser, **kwargs):
    allowed_keys = get_keys(section='processing')
    try:
        proc_sec = parser['PROCESSING']
    except KeyError:
        msg = "Section 'PROCESSING' does not exist in the config file"
        raise KeyError(msg)
    
    # override config file parameters with additional keyword arguments
    for k, v in kwargs.items():
        if k in allowed_keys:
            proc_sec[k] = v.strip()
    
    # make all relevant paths absolute
    for k in ['work_dir', 'scene_dir', 'scene', 'etad_dir']:
        v = proc_sec[k]
        proc_sec[k] = 'None' if v in ['', 'None'] else os.path.abspath(v)
    
    # set some defaults
    processing_defaults = {
        'sar_dir': 'SAR',
        'tmp_dir': 'TMP',
        'ard_dir': 'ARD',
        'wbm_dir': 'WBM',
        'gdal_threads': '4',
        'dem_type': 'Copernicus 30m Global DEM',
        'date_strict': 'True',
        'datatake': 'None',
        'measurement': 'gamma',
        'annotation': 'dm,ei,id,lc,li,np,ratio',
        'logfile': 'None',
        'parquet': 'None'
    }
    processing_options = {
        'acq_mode': ['IW', 'EW', 'SM'],
        'annotation': ['dm', 'ei', 'em', 'id', 'lc',
                       'ld', 'li', 'np', 'ratio', 'wm'],
        'dem_type': ['Copernicus 10m EEA DEM',
                     'Copernicus 30m Global DEM',
                     'Copernicus 30m Global DEM II',
                     'GETASSE30'],
        'measurement': ['gamma', 'sigma'],
        'mode': ['sar', 'nrb', 'orb'],
        'product': ['GRD', 'SLC'],
        'sensor': ['S1A', 'S1B', 'S1C', 'S1D']}
    
    if 'etad' not in proc_sec.keys():
        proc_sec['etad'] = 'False'
        proc_sec['etad_dir'] = 'None'
    for k, v in processing_defaults.items():
        if k not in proc_sec.keys():
            proc_sec[k] = v
    
    # check completeness of configuration parameters
    missing = []
    exclude = ['aoi_tiles', 'aoi_geometry']
    for key in get_keys(section='processing'):
        if key not in proc_sec.keys() and key not in exclude:
            missing.append(key)
    if len(missing) > 0:
        missing_str = '\n - ' + '\n - '.join(missing)
        raise RuntimeError(f"missing the following parameters:{missing_str}")
    
    out = {}
    for k, v in proc_sec.items():
        # check if key is allowed and convert 'None|none|' strings to None
        v = keyval_check(key=k, val=v, allowed_keys=allowed_keys)
        
        if k in ['annotation', 'aoi_tiles', 'data_take', 'mode', 'stac_collections']:
            v = proc_sec.get_list(k)
        
        validate_value(k, v)
        
        if k == 'mindate' and v is not None:
            v = proc_sec.get_datetime(k)
        if k == 'maxdate' and v is not None:
            date_short = re.search('^[0-9-]{10}$', v) is not None
            v = proc_sec.get_datetime(k)
            if date_short:
                v += timedelta(days=1, microseconds=-1)
        dir_ignore = ['work_dir']
        if proc_sec['etad'] == 'False':
            dir_ignore.append('etad_dir')
        if k == 'scene_dir' and v is None:
            dir_ignore.append(k)
        if k.endswith('_dir') and k not in dir_ignore:
            if os.path.isabs(v):
                msg = f"Parameter '{k}': '{v}' must be an existing directory"
                assert v is not None and os.path.isdir(v), msg
            else:
                v = os.path.join(proc_sec['work_dir'], v)
        if k.endswith('_file') and not k.startswith('db'):
            msg = f"Parameter '{k}': file {v} could not be found"
            if os.path.isabs(v):
                assert os.path.isfile(v), msg
            else:
                v = os.path.join(proc_sec['work_dir'], v)
                assert os.path.isfile(v), msg
        if k in ['db_file', 'logfile'] and v is not None:
            if not os.path.isabs(v):
                v = os.path.join(proc_sec['work_dir'], v)
        if k == 'gdal_threads':
            v = int(v)
        if k in ['etad', 'date_strict']:
            v = proc_sec.getboolean(k)
        
        validate_options(k, v, options=processing_options)
        out[k] = v
    
    # check that a valid scene search option is set
    db_file_set = out['db_file'] is not None
    stac_catalog_set = out['stac_catalog'] is not None
    stac_collections_set = out['stac_collections'] is not None
    parquet_set = out['parquet'] is not None
    
    options_set = sum([db_file_set, stac_catalog_set, parquet_set])
    
    if options_set == 0:
        raise RuntimeError("Please define a scene search option.")
    elif options_set > 1:
        raise RuntimeError("Multiple scene search options have been defined. Please choose only one.")
    
    if stac_catalog_set and not stac_collections_set:
        raise RuntimeError("'stac_collections' must be defined if data is to be searched in a STAC.")
    
    return out


def _parse_datetime(s):
    """Custom converter for configparser:
    https://docs.python.org/3/library/configparser.html#customizing-parser-behaviour"""
    return dateparse(s)


def _parse_list(s):
    """Custom converter for configparser:
    https://docs.python.org/3/library/configparser.html#customizing-parser-behaviour"""
    if s in ['', 'None']:
        return None
    else:
        return [x.strip() for x in s.split(',')]


def gdal_conf(config):
    """
    Stores GDAL configuration options for the current process.

    Parameters
    ----------
    config: dict
        Dictionary of the parsed config parameters for the current process.

    Returns
    -------
    dict
        Dictionary containing GDAL configuration options for the current process.
    """
    threads = config['processing']['gdal_threads']
    threads_before = gdal.GetConfigOption('GDAL_NUM_THREADS')
    if not isinstance(threads, int):
        raise TypeError("'threads' must be of type int")
    if threads == 1:
        multithread = False
    elif threads > 1:
        multithread = True
        gdal.SetConfigOption('GDAL_NUM_THREADS', str(threads))
    else:
        raise ValueError("'threads' must be >= 1")
    
    return {'threads': threads, 'threads_before': threads_before,
            'multithread': multithread}


def get_config(config_file: str | None = None, **kwargs: dict[str, str]) \
        -> dict[str, Any]:
    """
    Returns the content of a `config.ini` file as a dictionary.

    Parameters
    ----------
    config_file:
        Full path to the config file that should be parsed to a dictionary.
    kwargs:
        further keyword arguments overriding configuration found in the config file.

    Returns
    -------
        Dictionary of the parsed config parameters.
        The keys correspond to the config sections in lowercase letters.
    """
    parser = read_config_file(config_file)
    
    kwargs_proc = {k: v for k, v in kwargs.items() if k in get_keys('processing')}
    kwargs_meta = {k: v for k, v in kwargs.items() if k in get_keys('metadata')}
    
    out = {'processing': _get_config_processing(parser, **kwargs_proc),
           'metadata': _get_config_metadata(parser, **kwargs_meta)}
    
    processor_name = out['processing']['processor']
    processor = load_processor(processor_name)
    kwargs_sar = {k: v for k, v in kwargs.items() if k in get_keys(processor_name)}
    out[processor_name] = processor.get_config_section(parser, **kwargs_sar)
    
    return out


def get_keys(section: str) -> list[str]:
    """
    get all allowed configuration keys for a section
    
    Parameters
    ----------
    section:
        the configuration section to get the allowed keys for.
        Either 'processing', 'metadata' or the name of a SAR processor plugin e.g. 'snap'.

    Returns
    -------
        a list of keys
    """
    if section == 'processing':
        return ['acq_mode', 'annotation', 'aoi_geometry', 'aoi_tiles',
                'ard_dir', 'datatake', 'date_strict', 'db_file', 'dem_type',
                'etad', 'etad_dir', 'gdal_threads', 'logfile', 'maxdate',
                'measurement', 'mindate', 'mode', 'parquet', 'processor',
                'product', 'sar_dir', 'scene', 'scene_dir', 'sensor',
                'stac_catalog', 'stac_collections', 'tmp_dir', 'wbm_dir', 'work_dir']
    elif section == 'metadata':
        return ['access_url', 'copy_original', 'doi', 'format', 'licence', 'processing_center']
    else:
        try:
            processor = load_processor(section)
        except ModuleNotFoundError:
            raise RuntimeError(f"unknown section: {section}.")
        try:
            return processor.get_config_keys()
        except AttributeError:
            raise RuntimeError(f"missing function s1ard.{section}.get_config_keys().")


def init(
        target: str,
        source: str | None = None,
        overwrite: bool = False,
        **kwargs
) -> None:
    """
    Initialize a configuration file.

    Parameters
    ----------
    target:
        Path to the target configuration file.
    source:
        Path to the source file to read the configuration from. If not provided,
        a default configuration file within the package will be used.
    overwrite:
        Overwrite an existing file?
    kwargs:
        Additional keyword arguments for overwriting the configuration in `source`.

    Examples
    --------
    Create a file in the current working directory.
    `work_dir` and a scene search option (in this case SQLite via `db_file`)
    must be defined, other configuration is read from the default configuration file.

    >>> from s1ard.config import init
    >>> init(target='config.ini', work_dir='.', db_file='scenes.db')

    """
    if source is None:
        with importlib.resources.path(package='s1ard.resources',
                                      resource='config.ini') as path:
            source = str(path)
    config = get_config(config_file=source, **kwargs)
    write(config=config, target=target, overwrite=overwrite)


def read_config_file(config_file: str | None = None) -> configparser.ConfigParser:
    """
    Reads a configuration file and returns a ConfigParser object
    
    Parameters
    ----------
    config_file: str or None
        the configuration file name. If None, the default configuration file
        within the package will be used.

    Returns
    -------
        the configuration object
    """
    parser = configparser.ConfigParser(allow_no_value=True,
                                       converters={'_datetime': _parse_datetime,
                                                   '_list': _parse_list})
    
    if config_file:
        if not os.path.isfile(config_file):
            raise FileNotFoundError(f"Config file {config_file} does not exist.")
    else:
        with importlib.resources.path(package='s1ard.resources',
                                      resource='config.ini') as path:
            config_file = str(path)
    
    parser.read(config_file)
    return parser


def write(config, target, overwrite=False, **kwargs):
    """
    Write configuration options to a config file.
    
    Parameters
    ----------
    config: dict
        the configuration as returned by :func:`get_config`
    target: str
        the name of the output file
    overwrite: bool
        overwrite an existing file if it exists?
    kwargs
        further keyword arguments overriding configuration found in `config`.

    Returns
    -------

    """
    if os.path.isfile(target) and not overwrite:
        raise RuntimeError("target already exists")
    
    def to_string(item):
        """
        
        Parameters
        ----------
        item: dict or List or str

        Returns
        -------
        str or dict
        """
        if isinstance(item, dict):
            return {k: to_string(v) for k, v in item.items()}
        elif isinstance(item, list):
            return ', '.join([to_string(x) for x in item])
        elif isinstance(item, datetime):
            return item.strftime('%Y-%m-%d %H:%M:%S')
        else:
            return str(item)
    
    processor_name = config['processing']['processor']
    processor = load_processor(processor_name)
    
    config = copy.deepcopy(config)
    keys_processing = get_keys('processing')
    keys_meta = get_keys('metadata')
    keys_proc = processor.get_config_keys()
    for k, v in kwargs.items():
        if k in keys_processing:
            config['processing'][k] = v
        elif k in keys_meta:
            config['metadata'][k] = v
        elif k in keys_proc:
            config[processor_name][k] = v
        else:
            raise KeyError("Parameter '{}' is not supported".format(k))
    keys_path_relative = ['sar_dir', 'tmp_dir', 'ard_dir', 'wbm_dir', 'db_file']
    work_dir = config['processing']['work_dir']
    for k in keys_path_relative:
        v = config['processing'][k]
        if v is not None and work_dir in v:
            config['processing'][k] = v.replace(work_dir, '').strip('/\\')
    config['metadata'] = to_string(config['metadata'])
    config['processing'] = to_string(config['processing'])
    config_proc_str = processor.config_to_string(config[processor_name])
    config[processor_name] = config_proc_str
    parser = configparser.ConfigParser()
    parser['METADATA'] = config['metadata']
    parser['PROCESSING'] = config['processing']
    parser[processor_name.upper()] = config[processor_name]
    with open(target, 'w') as configfile:
        parser.write(configfile)



import os
import re
import sys
import shutil
import pystac
from spatialist.ancillary import finder
import logging

log = logging.getLogger('s1ard')


def make_catalog(directory, product_type, recursive=True, silent=False):
    """
    For a given directory of Sentinel-1 ARD products, this function will create a high-level STAC
    :class:`~pystac.catalog.Catalog` object serving as the STAC endpoint and lower-level STAC
    :class:`~pystac.collection.Collection` objects for each subdirectory corresponding to a unique MGRS tile ID.
    
    WARNING: The directory content will be reorganized into subdirectories based on the ARD type and unique MGRS tile
    IDs if this is not yet the case.
    
    Parameters
    ----------
    directory: str
        Path to a directory that contains ARD products.
    product_type: str
        Type of ARD products. Options: 'NRB' or 'ORB'.
    recursive: bool, optional
        Search `directory` recursively? Default is True.
    silent: bool, optional
        Should the output during directory reorganization be suppressed? Default is False.
    
    Returns
    -------
    nrb_catalog: pystac.catalog.Catalog
        STAC Catalog object
    
    Notes
    -----
    The returned STAC Catalog object contains Item asset hrefs that are absolute, whereas the actual on-disk files
    contain relative asset hrefs corresponding to the self-contained Catalog-Type. The returned in-memory STAC Catalog
    object deviates in this regard to ensure compatibility with the stackstac library:
    https://github.com/gjoseph92/stackstac/issues/20
    """
    overwrite = False
    product_type = product_type.upper()
    pattern = fr'^S1[AB]_(IW|EW|S[1-6])_{product_type}__1S(SH|SV|DH|DV|VV|HH|HV|VH)_[0-9]{{8}}T[0-9]{{6}}_[0-9]{{6}}_' \
              fr'[0-9A-F]{{6}}_[0-9A-Z]{{5}}_[0-9A-Z]{{4}}$'
    products = finder(target=directory, matchlist=[pattern], foldermode=2, regex=True, recursive=recursive)
    directory = os.path.join(directory, product_type)
    
    # Check if Catalog already exists
    catalog_path = os.path.join(directory, 'catalog.json')
    if os.path.isfile(catalog_path):
        overwrite = True
        catalog = pystac.Catalog.from_file(catalog_path)
        items = catalog.get_all_items()
        item_ids = [item.id for item in items]
        products_base = [os.path.basename(prod) for prod in products]
        diff = set(products_base) - set(item_ids)
        if len(diff) == 0:
            # See note in docstring - https://github.com/gjoseph92/stackstac/issues/20
            catalog.make_all_asset_hrefs_absolute()
            log.info(f"existing STAC endpoint found: {os.path.join(directory, 'catalog.json')}")
            return catalog
    
    sp_extent = pystac.SpatialExtent([None, None, None, None])
    tmp_extent = pystac.TemporalExtent([None, None])
    
    unique_tiles = list(
        set([re.search(re.compile(r'_[0-9A-Z]{5}_'), prod).group().replace('_', '') for prod in products]))
    products = _reorganize_by_tile(directory=directory, product_type=product_type, products=products,
                                   recursive=recursive, silent=silent)
    
    catalog = pystac.Catalog(id=f'{product_type.lower()}_catalog',
                             description=f'STAC Catalog of Sentinel-1 {product_type} products.',
                             title=f'STAC Catalog of Sentinel-1 {product_type} products.',
                             catalog_type=pystac.CatalogType.SELF_CONTAINED)
    
    for tile in unique_tiles:
        tile_collection = pystac.Collection(id=tile,
                                            description=f'STAC Collection of Sentinel-1 {product_type} products for '
                                                        f'MGRS tile {tile}.',
                                            title=f'STAC Collection of Sentinel-1 {product_type} products for '
                                                  f'MGRS tile {tile}.',
                                            extent=pystac.Extent(sp_extent, tmp_extent),
                                            keywords=['sar', 'backscatter', 'esa', 'copernicus', 'sentinel'],
                                            providers=[pystac.Provider(name='ESA',
                                                                       roles=[pystac.ProviderRole.LICENSOR,
                                                                              pystac.ProviderRole.PRODUCER])])
        catalog.add_child(tile_collection)
        
        items = []
        for prod in products:
            if tile in prod:
                item_path = os.path.join(prod, os.path.basename(prod) + '.json')
                item = pystac.read_file(href=item_path)
                items.append(item)
                tile_collection.add_item(item=item)
            else:
                continue
        
        extent = tile_collection.extent.from_items(items=items)
        tile_collection.extent = extent
    
    # Save Catalog and Collections on disk
    catalog.normalize_and_save(root_href=directory)
    
    # See note in docstring - https://github.com/gjoseph92/stackstac/issues/20
    catalog.make_all_asset_hrefs_absolute()
    
    if overwrite:
        log.info(f"existing STAC endpoint updated: {os.path.join(directory, 'catalog.json')}")
    else:
        log.info(f"new STAC endpoint created: {os.path.join(directory, 'catalog.json')}")
    return catalog


def _reorganize_by_tile(directory, product_type, products=None, recursive=True, silent=False):
    """
    Reorganizes a directory containing Sentinel-1 ARD products based on the ARD type and unique MGRS tile IDs.
    
    Parameters
    ----------
    directory: str
        Path to a directory that contains ARD products.
    product_type: str
        Type of ARD products. Options: 'NRB' or 'ORB'.
    products: list[str] or None, optional
        List of ARD product paths. Will be created from `directory` if not provided.
    recursive: bool, optional
        Search `directory` recursively? Default is True.
    silent: bool, optional
        If False (default), a message for each ARD product is printed if it has been moved to a new location or not.
    
    Returns
    -------
    products_new: list[str]
        An updated list of ARD product paths.
    """
    if products is None:
        parent_dir = os.path.dirname(directory)
        pattern = fr'^S1[AB]_(IW|EW|S[1-6])_{product_type}__1S(SH|SV|DH|DV|VV|HH|HV|VH)_[0-9]{{8}}T[0-9]{{6}}_' \
                  fr'[0-9]{{6}}_[0-9A-F]{{6}}_[0-9A-Z]{{5}}_[0-9A-Z]{{4}}$'
        products = finder(target=parent_dir, matchlist=[pattern], foldermode=2, regex=True, recursive=recursive)
    
    inp = input('WARNING:\n{}\nand the ARD products it contains will be reorganized into subdirectories '
                'based on unique MGRS tile IDs if this directory structure does not yet exist. '
                '\nDo you wish to continue? [yes|no] '.format(directory))
    if inp == 'yes':
        tile_dict = {}
        for prod in products:
            tile = re.search(re.compile(r'_[0-9A-Z]{5}_'), prod).group().replace('_', '')
            if tile in tile_dict and isinstance(tile_dict[tile], list):
                tile_dict[tile].append(prod)
            else:
                tile_dict[tile] = [prod]
        
        tiles = list(tile_dict.keys())
        products_new = []
        for tile in tiles:
            tile_dir = os.path.join(directory, tile)
            os.makedirs(tile_dir, exist_ok=True)
            
            for old_dir in tile_dict[tile]:
                new_dir = os.path.join(tile_dir, os.path.basename(old_dir))
                products_new.append(new_dir)
                
                if os.path.dirname(old_dir) != tile_dir:
                    shutil.move(old_dir, new_dir)
                    if not silent:
                        log.info(f"-> {os.path.basename(old_dir)} moved to {tile_dir}")
                else:
                    if not silent:
                        log.info(f"xx {os.path.basename(old_dir)} already in {tile_dir} (skip!)")
                    continue
        return products_new
    else:
        log.info('abort!')
        sys.exit(0)



from lxml import etree
import sys
import os


def validate(xml_path, xsd_path):
    message = ''
    try:
        xsd = etree.parse(xsd_path)
        xml = etree.parse(xml_path)
        schema = etree.XMLSchema(xsd)
        schema.validate(xml)
        message += str(schema.error_log)
    except Exception as e:
        message += str(e)
    finally:
        if message != '':
            print('{} is NOT VALID:\n{}'.format(xml_path, message))
            return 1
        else:
            print('{} is VALID!'.format(xml_path))
            return 0


if __name__ == '__main__':
    xml_path = sys.argv[1]
    xsd_path = sys.argv[2]
    if not xml_path.endswith('.xml') or not xsd_path.endswith('.xsd'):
        raise RuntimeError('Invalid input! Correct usage of this script:\n\n'
                           'python validate_xml.py your_xml_file.xml your_xsd_file.xsd\n')
    if any([not os.path.isfile(x) for x in [xml_path, xsd_path]]):
        raise FileNotFoundError('Input file(s) not found!')
    
    ret = validate(xml_path=xml_path, xsd_path=xsd_path)
    sys.exit(ret)
